"""Discrete transitions."""
from typing import Callable, Optional

import numpy as np

import probnum.random_variables as pnrv
from probnum.type import FloatArgType

from . import transition as trans


class DiscreteGaussian(trans.Transition):
    """Random variable transitions with additive Gaussian noise.

    .. math:: x_{i+1} \\sim \\mathcal{N}(g(t_i, x_i), S(t_i))

    for some (potentially non-linear) dynamics :math:`g` and diffusion matrix :math:`S`.
    This is used for, but not restricted to time-series.

    Parameters
    ----------
    dynamicsfun :
        Dynamics function :math:`g=g(t, x)`. Signature: ``dynafct(t, x)``.
    diffmatfun :
        Diffusion matrix function :math:`S=S(t)`. Signature: ``diffmatfct(t)``.
    jacobfun :
        Jacobian of the dynamics function :math:`g`, :math:`Jg=Jg(t, x)`.
        Signature: ``jacfct(t, x)``.

    See Also
    --------
    :class:`DiscreteModel`
    :class:`DiscreteGaussianLinearModel`
    """

    def __init__(
        self,
        dynamicsfun: Callable[[FloatArgType, np.ndarray], np.ndarray],
        diffmatfun: Callable[[FloatArgType], np.ndarray],
        jacobfun: Optional[Callable[[FloatArgType, np.ndarray], np.ndarray]] = None,
    ):
        self.dynamicsfun = dynamicsfun
        self.diffmatfun = diffmatfun

        def if_no_jacobian(t, x):
            raise NotImplementedError

        self.jacobfun = jacobfun if jacobfun is not None else if_no_jacobian
        super().__init__()

    def transition_realization(self, real, start, **kwargs):

        newmean = self.dynamicsfun(start, real)
        newcov = self.diffmatfun(start)
        crosscov = np.zeros(newcov.shape)
        return pnrv.Normal(newmean, newcov), {"crosscov": crosscov}

    def transition_rv(self, rv, start, **kwargs):

        raise NotImplementedError


class DiscreteLinearGaussian(DiscreteGaussian):
    """Discrete, linear Gaussian transition models of the form.

    .. math:: x_{i+1} \\sim \\mathcal{N}(G(t_i) x_i + v(t_i), S(t_i))

    for some dynamics matrix :math:`G=G(t)`, force vector :math:`v=v(t)`,
    and diffusion matrix :math:`S=S(t)`.


    Parameters
    ----------
    dynamatfct : callable
        Dynamics function :math:`G=G(t)`. Signature: ``dynamatfct(t)``.
    forcefct : callable
        Force function :math:`v=v(t)`. Signature: ``forcefct(t)``.
    diffmatfct : callable
        Diffusion matrix function :math:`S=S(t)`. Signature: ``diffmatfct(t)``.

    See Also
    --------
    :class:`DiscreteModel`
    :class:`DiscreteGaussianLinearModel`
    """

    def __init__(
        self,
        dynamicsmatfun: Callable[[FloatArgType], np.ndarray],
        forcevecfun: Callable[[FloatArgType], np.ndarray],
        diffmatfun: Callable[[FloatArgType], np.ndarray],
    ):

        self.dynamicsmatfun = dynamicsmatfun
        self.forcevecfun = forcevecfun

        super().__init__(
            dynamicsfun=lambda t, x: (self.dynamicsmatfun(t) @ x + self.forcevecfun(t)),
            diffmatfun=diffmatfun,
            jacobfun=lambda t, x: dynamicsmatfun(t),
        )

    def transition_rv(self, rv, start, **kwargs):

        if not isinstance(rv, pnrv.Normal):
            raise TypeError(f"Normal RV expected, but {type(rv)} received.")

        dynamicsmat = self.dynamicsmatfun(start)
        diffmat = self.diffmatfun(start)
        force = self.forcevecfun(start)

        new_mean = dynamicsmat @ rv.mean + force
        new_crosscov = rv.cov @ dynamicsmat.T
        new_cov = dynamicsmat @ new_crosscov + diffmat
        return pnrv.Normal(mean=new_mean, cov=new_cov), {"crosscov": new_crosscov}

    @property
    def dimension(self):
        # risky to evaluate at zero, but works well
        # remove this -- the dimension of discrete transitions is not clear!
        # input dim != output dim is possible...
        # See issue #266
        return len(self.dynamicsmatfun(0.0).T)


class DiscreteLTIGaussian(DiscreteLinearGaussian):
    """Discrete, linear, time-invariant Gaussian transition models of the form.

    .. math:: x_{i+1} \\sim \\mathcal{N}(G x_i + v, S)

    for some dynamics matrix :math:`G`, force vector :math:`v`,
    and diffusion matrix :math:`S`.

    Parameters
    ----------
    dynamat :
        Dynamics matrix :math:`G`.
    forcevec :
        Force vector :math:`v`.
    diffmat :
        Diffusion matrix :math:`S`.

    Raises
    ------
    TypeError
        If dynamat, forcevec and diffmat have incompatible shapes.

    See Also
    --------
    :class:`DiscreteModel`
    :class:`DiscreteGaussianLinearModel`
    """

    def __init__(
        self, dynamicsmat: np.ndarray, forcevec: np.ndarray, diffmat: np.ndarray
    ):
        _check_dimensions(dynamicsmat, forcevec, diffmat)

        super().__init__(
            lambda t: dynamicsmat,
            lambda t: forcevec,
            lambda t: diffmat,
        )

        self.dynamicsmat = dynamicsmat
        self.forcevec = forcevec
        self.diffmat = diffmat


def _check_dimensions(dynamicsmat, forcevec, diffmat):
    if dynamicsmat.ndim != 2:
        raise TypeError(
            f"dynamat.ndim=2 expected. dynamat.ndim={dynamicsmat.ndim} received."
        )
    if forcevec.ndim != 1:
        raise TypeError(
            f"forcevec.ndim=1 expected. forcevec.ndim={forcevec.ndim} received."
        )
    if diffmat.ndim != 2:
        raise TypeError(
            f"diffmat.ndim=2 expected. diffmat.ndim={diffmat.ndim} received."
        )
    if (
        dynamicsmat.shape[0] != forcevec.shape[0]
        or forcevec.shape[0] != diffmat.shape[0]
        or diffmat.shape[0] != diffmat.shape[1]
    ):
        raise TypeError(
            f"Dimension of dynamat, forcevec and diffmat do not align. "
            f"Expected: dynamat.shape=(N,*), forcevec.shape=(N,), diffmat.shape=(N, N).     "
            f"Received: dynamat.shape={dynamicsmat.shape}, forcevec.shape={forcevec.shape}, "
            f"diffmat.shape={diffmat.shape}."
        )



from typing import Union

import numpy as np

import probnum as pn


class _RandomVariableList(list):
    """List of RandomVariables with convenient access to means, covariances, etc.

    Parameters
    ----------
    rv_list :
        :obj:`list` of :obj:`RandomVariable`
    """

    def __init__(self, rv_list: list):
        if not isinstance(rv_list, list):
            raise TypeError("RandomVariableList expects a list.")

        # First element as a proxy for checking all elements
        if not isinstance(rv_list[0], pn.RandomVariable):
            raise TypeError(
                "RandomVariableList expects RandomVariable elements, but "
                + f"first element has type {type(rv_list[0])}."
            )
        super().__init__(rv_list)

    @property
    def mean(self) -> np.ndarray:
        return np.stack([rv.mean for rv in self])

    @property
    def cov(self) -> np.ndarray:
        return np.stack([rv.cov for rv in self])

    @property
    def var(self) -> np.ndarray:
        return np.stack([rv.var for rv in self])

    @property
    def std(self) -> np.ndarray:
        return np.stack([rv.std for rv in self])

    def __getitem__(self, idx) -> Union["pn.RandomVariable", list]:
        result = super().__getitem__(idx)
        # Make sure to wrap the result into a _RandomVariableList if necessary
        if isinstance(result, list):
            result = _RandomVariableList(result)
        return result



"""Interfaces for Bayesian filtering and smoothing."""

from abc import ABC, abstractmethod


class BayesFiltSmooth(ABC):
    """Bayesian filtering and smoothing."""

    def __init__(self, dynamics_model, measurement_model, initrv):
        self.dynamics_model = dynamics_model
        self.measurement_model = measurement_model
        self.initrv = initrv

    @abstractmethod
    def filter_step(self, start, stop, randvar, data, **kwargs):
        """Filter step.

        For e.g. Gaussian filters, this means a prediction step followed
        by an update step.
        """
        errormsg = (
            "filter_step(...) is not implemented for "
            + "the Bayesian filter {}.".format(type(self).__name__)
        )
        raise NotImplementedError(errormsg)

    def smoother_step(self, **kwargs):
        """Smoother step."""
        errormsg = (
            "smoother_step(...) is not implemented for "
            + "the Bayesian smoother {}.".format(type(self).__name__)
        )
        raise NotImplementedError(errormsg)



"""Import convenience functions in optim.py to create an intuitive, numpy-like
interface.

Note
----
Local import, because with a global import this does not seem
to work.
"""

from .ivp2filter import ivp2ekf0, ivp2ekf1, ivp2ukf
from .ivpfiltsmooth import GaussianIVPFilter
from .odefiltsmooth import probsolve_ivp



"""General Gaussian filters based on approximating intractable quantities with numerical
quadrature.

Examples include the unscented Kalman filter / RTS smoother which is
based on a third degree fully symmetric rule.
"""

import numpy as np

from probnum.filtsmooth import statespace
from probnum.filtsmooth.gaussfiltsmooth import unscentedtransform as ut
from probnum.random_variables import Normal


class ContinuousUKFComponent(statespace.Transition):
    """Continuous unscented Kalman filter transition."""

    def __init__(
        self, non_linear_sde, dimension, spread=1e-4, priorpar=2.0, special_scale=0.0
    ):
        if not isinstance(non_linear_sde, statespace.SDE):
            raise TypeError("cont_model must be an SDE.")
        self.non_linear_sde = non_linear_sde
        self.ut = ut.UnscentedTransform(dimension, spread, priorpar, special_scale)
        super().__init__()

        raise NotImplementedError("Implementation incomplete.")

    def transition_realization(self, real, start, stop, linearise_at=None, **kwargs):
        raise NotImplementedError("TODO")  # Issue  #234

    def transition_rv(self, rv, start, stop, linearise_at=None, **kwargs):
        raise NotImplementedError("TODO")  # Issue  #234

    @property
    def dimension(self):
        raise NotImplementedError


class DiscreteUKFComponent(statespace.Transition):
    """Discrete extended Kalman filter transition."""

    def __init__(
        self, disc_model, dimension, spread=1.0, priorpar=2.0, special_scale=0.0
    ):
        self.disc_model = disc_model
        self.ut = ut.UnscentedTransform(dimension, spread, priorpar, special_scale)
        super().__init__()

    def transition_realization(self, real, start, **kwargs):
        return self.disc_model.transition_realization(real, start, **kwargs)

    def transition_rv(self, rv, start, linearise_at=None, **kwargs):
        compute_sigmapts_at = linearise_at if linearise_at is not None else rv
        sigmapts = self.ut.sigma_points(
            compute_sigmapts_at.mean, compute_sigmapts_at.cov
        )
        proppts = self.ut.propagate(start, sigmapts, self.disc_model.dynamicsfun)
        meascov = self.disc_model.diffmatfun(start)
        mean, cov, crosscov = self.ut.estimate_statistics(
            proppts, sigmapts, meascov, rv.mean
        )
        return Normal(mean, cov), {"crosscov": crosscov}

    @property
    def dimension(self):
        return self.ut.dimension

    @classmethod
    def from_ode(cls, ode, prior, evlvar):

        spatialdim = prior.spatialdim
        h0 = prior.proj2coord(coord=0)
        h1 = prior.proj2coord(coord=1)

        def dyna(t, x):
            return h1 @ x - ode.rhs(t, h0 @ x)

        def diff(t):
            return evlvar * np.eye(spatialdim)

        disc_model = statespace.DiscreteGaussian(dyna, diff)
        return cls(disc_model, dimension=prior.dimension)



"""SDE models as transitions."""
import functools
from typing import Callable

import numpy as np
import scipy.linalg

import probnum.random_variables as pnrv
from probnum.type import FloatArgType

from . import discrete_transition, transition


class SDE(transition.Transition):
    """Stochastic differential equation.

    .. math:: d x(t) = g(t, x(t)) d t + L(t) d w(t),

    driven by a Wiener process with unit diffusion.
    """

    def __init__(
        self,
        driftfun: Callable[[FloatArgType, np.ndarray], np.ndarray],
        dispmatfun: Callable[[FloatArgType, np.ndarray], np.ndarray],
        jacobfun: Callable[[FloatArgType, np.ndarray], np.ndarray],
    ):
        self.driftfun = driftfun
        self.dispmatfun = dispmatfun
        self.jacobfun = jacobfun
        super().__init__()

    def transition_realization(
        self,
        real,
        start,
        stop=None,
        step=None,
        linearise_at=None,
    ):
        raise NotImplementedError

    def transition_rv(
        self,
        rv,
        start,
        stop=None,
        step=None,
        linearise_at=None,
    ):
        raise NotImplementedError

    @property
    def dimension(self):
        raise NotImplementedError


class LinearSDE(SDE):
    """Linear stochastic differential equation (SDE),

    .. math:: d x(t) = [G(t) x(t) + v(t)] d t + L(t) x(t) d w(t).

    For Gaussian initial conditions, this solution is a Gaussian process.

    Parameters
    ----------
    driftmatfun :
        This is G = G(t). The evaluations of this function are called
        the driftmatrix of the SDE.
        Returns np.ndarray with shape=(n, n)
    forcevecfun :
        This is v = v(t). Evaluations of this function are called
        the force(vector) of the SDE.
        Returns np.ndarray with shape=(n,)
    dispmatfun :
        This is L = L(t). Evaluations of this function are called
        the dispersion(matrix) of the SDE.
        Returns np.ndarray with shape=(n, s)
    """

    def __init__(
        self,
        driftmatfun: Callable[[FloatArgType], np.ndarray],
        forcevecfun: Callable[[FloatArgType], np.ndarray],
        dispmatfun: Callable[[FloatArgType], np.ndarray],
    ):
        self.driftmatfun = driftmatfun
        self.forcevecfun = forcevecfun
        super().__init__(
            driftfun=(lambda t, x: driftmatfun(t) @ x + forcevecfun(t)),
            dispmatfun=dispmatfun,
            jacobfun=(lambda t, x: driftmatfun(t)),
        )

    def transition_realization(
        self,
        real,
        start,
        stop,
        step,
        **kwargs,
    ):

        rv = pnrv.Normal(real, 0 * np.eye(len(real)))
        return linear_sde_statistics(
            rv,
            start,
            stop,
            step,
            self.driftfun,
            self.driftmatfun,
            self.dispmatfun,
        )

    def transition_rv(self, rv, start, stop, step, **kwargs):

        if not isinstance(rv, pnrv.Normal):
            errormsg = (
                "Closed form transitions in linear SDE models is only "
                "available for Gaussian initial conditions."
            )
            raise TypeError(errormsg)
        return linear_sde_statistics(
            rv,
            start,
            stop,
            step,
            self.driftfun,
            self.driftmatfun,
            self.dispmatfun,
        )

    @property
    def dimension(self):
        """Spatial dimension (utility attribute)."""
        # risky to evaluate at zero, but usually works
        return len(self.driftmatfun(0.0))


class LTISDE(LinearSDE):
    """Linear time-invariant continuous Markov models of the form.

    .. math:: d x(t) = [G x(t) + v] d t + L d w(t).

    In the language of dynamic models,
    x(t) : state process
    G : drift matrix
    v : force term/vector
    L : dispersion matrix.
    w(t) : Wiener process with unit diffusion.

    Parameters
    ----------
    driftmat :
        This is F. It is the drift matrix of the SDE.
    forcevec :
        This is U. It is the force vector of the SDE.
    dispmat :
        This is L. It is the dispersion matrix of the SDE.
    """

    def __init__(self, driftmat: np.ndarray, forcevec: np.ndarray, dispmat: np.ndarray):
        _check_initial_state_dimensions(driftmat, forcevec, dispmat)
        super().__init__(
            (lambda t: driftmat),
            (lambda t: forcevec),
            (lambda t: dispmat),
        )
        self.driftmat = driftmat
        self.forcevec = forcevec
        self.dispmat = dispmat

    def transition_realization(
        self,
        real,
        start,
        stop,
        **kwargs,
    ):

        if not isinstance(real, np.ndarray):
            raise TypeError(f"Numpy array expected, {type(real)} received.")
        discretised_model = self.discretise(step=stop - start)
        return discretised_model.transition_realization(real, start)

    def transition_rv(
        self,
        rv,
        start,
        stop,
        **kwargs,
    ):

        if not isinstance(rv, pnrv.Normal):
            errormsg = (
                "Closed form transitions in LTI SDE models is only "
                "available for Gaussian initial conditions."
            )
            raise TypeError(errormsg)
        discretised_model = self.discretise(step=stop - start)
        return discretised_model.transition_rv(rv, start)

    def discretise(self, step):
        """Returns a discrete transition model (i.e. mild solution to SDE) using matrix
        fraction decomposition.

        That is, matrices A(h) and Q(h) and vector s(h) such
        that the transition is

        .. math:: x | x_\\text{old} \\sim \\mathcal{N}(A(h) x_\\text{old} + s(h), Q(h)) ,

        which is the transition of the mild solution to the LTI SDE.
        """
        if np.linalg.norm(self.forcevec) > 0:
            raise NotImplementedError("MFD does not work for force>0 (yet).")
        ah, qh, _ = matrix_fraction_decomposition(self.driftmat, self.dispmat, step)
        sh = np.zeros(len(ah))
        return discrete_transition.DiscreteLTIGaussian(ah, sh, qh)


def linear_sde_statistics(rv, start, stop, step, driftfun, jacobfun, dispmatfun):
    """Computes mean and covariance of SDE solution.

    For a linear(ised) SDE

    .. math:: d x(t) = [G(t) x(t) + v(t)] d t + L(t) x(t) d w(t).

    mean and covariance of the solution are computed by solving

    .. math:: \\frac{dm}{dt}(t) = G(t) m(t) + v(t), \\frac{dC}{dt}(t) = G(t) C(t) + C(t) G(t)^\\top + L(t) L(t)^\\top,

    which is done here with a few steps of the RK4 method.
    This function is also called by the continuous-time extended Kalman filter,
    which is why the drift can be any function.

    Parameters
    ----------
    rv :
        Normal random variable. Distribution of mean and covariance at the starting point.
    start :
        Start of the time-interval
    stop :
        End of the time-interval
    step :
        Step-size used in RK4.
    driftfun :
        Drift of the (non)linear SDE
    jacobfun :
        Jacobian of the drift function
    dispmatfun :
        Dispersion matrix function

    Returns
    -------
    Normal random variable
        Mean and covariance are the solution of the differential equation
    dict
        Empty dict, may be extended in the future to contain information
        about the solution process, e.g. number of function evaluations.
    """
    if step <= 0.0:
        raise ValueError("Step-size must be positive.")
    mean, cov = rv.mean, rv.cov
    time = start

    # Set default arguments for frequently used functions.
    increment_fun = functools.partial(
        _increment_fun,
        driftfun=driftfun,
        jacobfun=jacobfun,
        dispmatfun=dispmatfun,
    )
    rk4_step = functools.partial(_rk4_step, step=step, fun=increment_fun)

    while time < stop:
        mean, cov, time = rk4_step(mean, cov, time)
    return pnrv.Normal(mean, cov), {}


def _rk4_step(mean, cov, time, step, fun):
    """Do a single RK4 step to compute the solution."""
    m1, c1 = fun(time, mean, cov)
    m2, c2 = fun(time + step / 2.0, mean + step * m1 / 2.0, cov + step * c1 / 2.0)
    m3, c3 = fun(time + step / 2.0, mean + step * m2 / 2.0, cov + step * c2 / 2.0)
    m4, c4 = fun(time + step, mean + step * m3, cov + step * c3)
    mean = mean + step * (m1 + 2 * m2 + 2 * m3 + m4) / 6.0
    cov = cov + step * (c1 + 2 * c2 + 2 * c3 + c4) / 6.0
    time = time + step
    return mean, cov, time


def _increment_fun(time, mean, cov, driftfun, jacobfun, dispmatfun):
    """Euler step for closed form solutions of ODE defining mean and covariance of the
    closed-form transition.

    Maybe make this into a different solver (euler sucks).

    See RHS of Eq. 10.82 in Applied SDEs.
    """
    dispersion_matrix = dispmatfun(time)
    jacobian = jacobfun(time)
    mean_increment = driftfun(time, mean)
    cov_increment = (
        cov @ jacobian.T + jacobian @ cov.T + dispersion_matrix @ dispersion_matrix.T
    )
    return mean_increment, cov_increment


def matrix_fraction_decomposition(driftmat, dispmat, step):
    """Matrix fraction decomposition (without force)."""
    no_force = np.zeros(len(driftmat))
    _check_initial_state_dimensions(
        driftmat=driftmat, forcevec=no_force, dispmat=dispmat
    )

    topleft = driftmat
    topright = dispmat @ dispmat.T
    bottomright = -driftmat.T
    bottomleft = np.zeros(driftmat.shape)

    toprow = np.hstack((topleft, topright))
    bottomrow = np.hstack((bottomleft, bottomright))
    bigmat = np.vstack((toprow, bottomrow))

    Phi = scipy.linalg.expm(bigmat * step)
    projmat1 = np.eye(*toprow.shape)
    projmat2 = np.flip(projmat1)

    Ah = projmat1 @ Phi @ projmat1.T
    C, D = projmat1 @ Phi @ projmat2.T, Ah.T
    Qh = C @ D

    return Ah, Qh, bigmat


def _check_initial_state_dimensions(driftmat, forcevec, dispmat):
    """Checks that the matrices all align and are of proper shape.

    Parameters
    ----------
    driftmat : np.ndarray, shape=(n, n)
    forcevec : np.ndarray, shape=(n,)
    dispmat : np.ndarray, shape=(n, s)
    """
    if driftmat.ndim != 2 or driftmat.shape[0] != driftmat.shape[1]:
        raise ValueError("driftmatrix not of shape (n, n)")
    if forcevec.ndim != 1:
        raise ValueError("force not of shape (n,)")
    if forcevec.shape[0] != driftmat.shape[1]:
        raise ValueError("force not of shape (n,) or driftmatrix not of shape (n, n)")
    if dispmat.ndim != 2:
        raise ValueError("dispersion not of shape (n, s)")



"""Gaussian filtering and smoothing."""

import numpy as np

from probnum._randomvariablelist import _RandomVariableList
from probnum.filtsmooth.bayesfiltsmooth import BayesFiltSmooth
from probnum.filtsmooth.gaussfiltsmooth.kalmanposterior import KalmanPosterior
from probnum.random_variables import Normal


class Kalman(BayesFiltSmooth):
    """Gaussian filtering and smoothing, i.e. Kalman-like filters and smoothers."""

    def __init__(self, dynamics_model, measurement_model, initrv):
        """Check that the initial distribution is Gaussian."""
        if not issubclass(type(initrv), Normal):
            raise ValueError(
                "Gaussian filters/smoothers need initial "
                "random variables with Normal distribution."
            )
        super().__init__(dynamics_model, measurement_model, initrv)

    def filtsmooth(self, dataset, times, intermediate_step=None):
        """Apply Gaussian filtering and smoothing to a data set.

        Parameters
        ----------
        dataset : array_like, shape (N, M)
            Data set that is filtered.
        times : array_like, shape (N,)
            Temporal locations of the data points.

        Returns
        -------
        KalmanPosterior
            Posterior distribution of the smoothed output
        """
        dataset, times = np.asarray(dataset), np.asarray(times)
        filter_posterior = self.filter(
            dataset, times, intermediate_step=intermediate_step
        )
        smooth_posterior = self.smooth(
            filter_posterior, intermediate_step=intermediate_step
        )
        return smooth_posterior

    def filter(self, dataset, times, intermediate_step=None, linearise_at=None):
        """Apply Gaussian filtering (no smoothing!) to a data set.

        Parameters
        ----------
        dataset : array_like, shape (N, M)
            Data set that is filtered.
        times : array_like, shape (N,)
            Temporal locations of the data points.

        Returns
        -------
        KalmanPosterior
            Posterior distribution of the filtered output
        """
        # linearise_at is not used here, only in IteratedKalman.filter_step
        # which is overwritten by IteratedKalman
        dataset, times = np.asarray(dataset), np.asarray(times)
        filtrv = self.initrv
        rvs = [filtrv]
        for idx in range(1, len(times)):
            filtrv, _ = self.filter_step(
                start=times[idx - 1],
                stop=times[idx],
                current_rv=filtrv,
                data=dataset[idx - 1],
                intermediate_step=intermediate_step,
            )
            rvs.append(filtrv)
        return KalmanPosterior(times, rvs, self, with_smoothing=False)

    def filter_step(self, start, stop, current_rv, data, intermediate_step=None):
        """A single filter step.

        Consists of a prediction step (t -> t+1) and an update step (at t+1).

        Parameters
        ----------
        start : float
            Predict FROM this time point.
        stop : float
            Predict TO this time point.
        current_rv : RandomVariable
            Predict based on this random variable. For instance, this can be the result
            of a previous call to filter_step.
        data : array_like
            Compute the update based on this data.

        Returns
        -------
        RandomVariable
            Resulting filter estimate after the single step.
        dict
            Additional information provided by predict() and update().
            Contains keys `pred_rv`, `info_pred`, `meas_rv`, `info_upd`.
        """
        data = np.asarray(data)
        info = {}
        info["pred_rv"], info["info_pred"] = self.predict(
            start, stop, current_rv, intermediate_step=intermediate_step
        )
        filtrv, info["meas_rv"], info["info_upd"] = self.update(
            stop, info["pred_rv"], data
        )
        return filtrv, info

    def predict(self, start, stop, randvar, intermediate_step=None):
        return self.dynamics_model.transition_rv(
            randvar,
            start,
            stop=stop,
            step=intermediate_step,
        )

    def measure(self, time, randvar):
        """Propagate the state through the measurement model.

        Parameters
        ----------
        time : float
            Time of the measurement.
        randvar : Normal
            Random variable to be propagated through the measurement model.

        Returns
        -------
        meas_rv : Normal
            Measured random variable, as returned by the measurement model.
        info : dict
            Additional info. Contains at leas the key `crosscov` which is the cross
            covariance between the input random variable and the measured random
            variable.
        """
        return self.measurement_model.transition_rv(randvar, time)

    def condition_state_on_measurement(self, randvar, meas_rv, data, crosscov):
        """Condition the state on the observed data.

        Parameters
        ----------
        randvar : Normal
            Random variable to be updated with the measurement and data.
        meas_rv : Normal
            Measured random variable, as returned by the measurement model.
        data : np.ndarray
            Data to update on.
        crosscov : np.ndarray
            Cross-covariance between the state random variable `randvar` and the
            measurement random variable `meas_rv`.

        Returns
        -------
        Normal
            Updated Normal random variable (new filter estimate)
        """
        new_mean = randvar.mean + crosscov @ np.linalg.solve(
            meas_rv.cov, data - meas_rv.mean
        )
        new_cov = randvar.cov - crosscov @ np.linalg.solve(meas_rv.cov, crosscov.T)
        filt_rv = Normal(new_mean, new_cov)
        return filt_rv

    def update(self, time, randvar, data):
        """Gaussian filter update step. Consists of a measurement step and a
        conditioning step.

        Parameters
        ----------
        time : float
            Time of the update.
        randvar : RandomVariable
            Random variable to be updated. Result of :meth:`predict()`.
        data : np.ndarray
            Data to update on.

        Returns
        -------
        filt_rv : Normal
            Updated Normal RV (new filter estimate).
        meas_rv : Normal
            Measured random variable, as returned by the measurement model.
        info : dict
            Additional info. Contains at least the key `crosscov`,
            which is the crosscov between input RV and measured RV.
            The crosscov does not relate to the updated RV!
        """
        meas_rv, info = self.measure(time, randvar)
        filt_rv = self.condition_state_on_measurement(
            randvar, meas_rv, data, info["crosscov"]
        )
        return filt_rv, meas_rv, info

    def smooth(self, filter_posterior, intermediate_step=None):
        """Apply Gaussian smoothing to the filtering outcome (i.e. a KalmanPosterior).

        Parameters
        ----------
        filter_posterior : KalmanPosterior
            Posterior distribution obtained after filtering
        intermediate_step :
            Step-size to be taken by approximate transition methods.
        Returns
        -------
        KalmanPosterior
            Posterior distribution of the smoothed output
        """
        rv_list = self.smooth_list(
            filter_posterior,
            filter_posterior.locations,
            intermediate_step=intermediate_step,
        )
        return KalmanPosterior(
            filter_posterior.locations, rv_list, self, with_smoothing=True
        )

    def smooth_list(self, rv_list, locations, intermediate_step=None):
        """Apply smoothing to a list of RVs with desired final random variable.

        Specification of a final RV is useful to compute joint samples from a KalmanPosterior object,
        because in this case, the final RV is a Dirac (over a sample from the final Normal RV)
        and not a Normal RV.

        Parameters
        ----------
        rv_list : _RandomVariableList or array_like
            List of random variables to be smoothed.
        locations : array_like
            Locations of the random variables in rv_list.
        intermediate_step :
            Step-size to be taken by approximate transition methods.

        Returns
        -------
        _RandomVariableList
            List of smoothed random variables.
        """
        final_rv = rv_list[-1]
        curr_rv = final_rv
        out_rvs = [curr_rv]
        for idx in reversed(range(1, len(locations))):
            unsmoothed_rv = rv_list[idx - 1]
            curr_rv = self.smooth_step(
                unsmoothed_rv,
                curr_rv,
                start=locations[idx - 1],
                stop=locations[idx],
                intermediate_step=intermediate_step,
            )
            out_rvs.append(curr_rv)
        out_rvs.reverse()
        return _RandomVariableList(out_rvs)

    def smooth_step(
        self, unsmoothed_rv, smoothed_rv, start, stop, intermediate_step=None
    ):
        """A single smoother step.

        Consists of predicting from the filtering distribution at time t
        to time t+1 and then updating based on the discrepancy to the
        smoothing solution at time t+1.
        If preconditioning is available in the dynamic model, this is leveraged here.
        If not, a classic smoothing step estimate is taken.

        Parameters
        ----------
        unsmoothed_rv : RandomVariable
            Filtering distribution at time t.
        smoothed_rv : RandomVariable
            Prediction at time t+1 of the filtering distribution at time t.
        start : float
            Time-point of the to-be-smoothed RV.
        stop : float
            Time-point of the already-smoothed RV.
        intermediate_step :
            Step-size to be taken by approximate transition methods.
        """
        if self.dynamics_model.precon is None:
            return self._smooth_step_classic(
                unsmoothed_rv,
                smoothed_rv,
                start,
                stop,
                intermediate_step=intermediate_step,
            )
        else:
            return self._smooth_step_with_preconditioning(
                unsmoothed_rv,
                smoothed_rv,
                start,
                stop,
                intermediate_step=intermediate_step,
            )

    def _smooth_step_classic(
        self, unsmoothed_rv, smoothed_rv, start, stop, intermediate_step=None
    ):
        """A single smoother step.

        Consists of predicting from the filtering distribution at time t
        to time t+1 and then updating based on the discrepancy to the
        smoothing solution at time t+1.

        Parameters
        ----------
        unsmoothed_rv : RandomVariable
            Filtering distribution at time t.
        smoothed_rv : RandomVariable
            Prediction at time t+1 of the filtering distribution at time t.
        start : float
            Time-point of the to-be-smoothed RV.
        stop : float
            Time-point of the already-smoothed RV.
        """
        predicted_rv, info = self.dynamics_model.transition_rv(
            unsmoothed_rv, start, stop=stop, step=intermediate_step
        )
        crosscov = info["crosscov"]
        smoothing_gain = np.linalg.solve(predicted_rv.cov.T, crosscov.T).T
        new_mean = unsmoothed_rv.mean + smoothing_gain @ (
            smoothed_rv.mean - predicted_rv.mean
        )
        new_cov = (
            unsmoothed_rv.cov
            + smoothing_gain @ (smoothed_rv.cov - predicted_rv.cov) @ smoothing_gain.T
        )
        return Normal(new_mean, new_cov)

    def _smooth_step_with_preconditioning(
        self, unsmoothed_rv, smoothed_rv, start, stop, intermediate_step=None
    ):
        """A single smoother step.

        Consists of predicting from the filtering distribution at time t
        to time t+1 and then updating based on the discrepancy to the
        smoothing solution at time t+1.

        Parameters
        ----------
        unsmoothed_rv : RandomVariable
            Filtering distribution at time t.
        smoothed_rv : RandomVariable
            Prediction at time t+1 of the filtering distribution at time t.
        start : float
            Time-point of the to-be-smoothed RV.
        stop : float
            Time-point of the already-smoothed RV.
        """
        # It is not clear to me how to best test this, except running IBM smoothing for high-ish order. (N)
        precon_inv = self.dynamics_model.precon.inverse(stop - start)
        unsmoothed_rv = precon_inv @ unsmoothed_rv
        smoothed_rv = precon_inv @ smoothed_rv

        predicted_rv, info = self.dynamics_model.transition_rv_preconditioned(
            unsmoothed_rv, start, stop=stop, step=intermediate_step
        )
        crosscov = info["crosscov"]
        smoothing_gain = np.linalg.solve(predicted_rv.cov.T, crosscov.T).T
        new_mean = unsmoothed_rv.mean + smoothing_gain @ (
            smoothed_rv.mean - predicted_rv.mean
        )
        new_cov = (
            unsmoothed_rv.cov
            + smoothing_gain @ (smoothed_rv.cov - predicted_rv.cov) @ smoothing_gain.T
        )
        return self.dynamics_model.precon(stop - start) @ Normal(new_mean, new_cov)



"""Gaussian filtering and smoothing based on making intractable quantities tractable
through Taylor-method approximations, e.g. linearization."""

import numpy as np

import probnum.random_variables as pnrv
from probnum.filtsmooth import statespace


class ContinuousEKFComponent(statespace.Transition):
    """Continuous extended Kalman filter transition."""

    def __init__(self, non_linear_sde, num_steps):
        if not isinstance(non_linear_sde, statespace.SDE):
            raise TypeError("Continuous EKF transition requires a (non-linear) SDE.")
        self.non_linear_sde = non_linear_sde
        self.num_steps = num_steps
        super().__init__()

    def transition_realization(self, real, start, stop, linearise_at=None, **kwargs):

        compute_jacobian_at = linearise_at.mean if linearise_at is not None else real

        def jacobfun(t, x=compute_jacobian_at):
            # replaces functools (second variable may not be called x)
            return self.non_linear_sde.jacobfun(t, x)

        step = (stop - start) / self.num_steps
        return statespace.linear_sde_statistics(
            rv=pnrv.Normal(mean=real, cov=np.zeros((len(real), len(real)))),
            start=start,
            stop=stop,
            step=step,
            driftfun=self.non_linear_sde.driftfun,
            jacobfun=jacobfun,
            dispmatfun=self.non_linear_sde.dispmatfun,
        )

    def transition_rv(self, rv, start, stop, linearise_at=None, **kwargs):

        compute_jacobian_at = linearise_at.mean if linearise_at is not None else rv.mean

        def jacobfun(t, x=compute_jacobian_at):
            # replaces functools (second variable may not be called x)
            return self.non_linear_sde.jacobfun(t, x)

        step = (stop - start) / self.num_steps
        return statespace.linear_sde_statistics(
            rv=rv,
            start=start,
            stop=stop,
            step=step,
            driftfun=self.non_linear_sde.driftfun,
            jacobfun=jacobfun,
            dispmatfun=self.non_linear_sde.dispmatfun,
        )

    @property
    def dimension(self):
        raise NotImplementedError


class DiscreteEKFComponent(statespace.Transition):
    """Discrete extended Kalman filter transition."""

    def __init__(self, disc_model):
        self.disc_model = disc_model
        super().__init__()

    def transition_realization(self, real, start, **kwargs):
        return self.disc_model.transition_realization(real, start, **kwargs)

    def transition_rv(self, rv, start, linearise_at=None, **kwargs):
        diffmat = self.disc_model.diffmatfun(start)
        compute_jacobian_at = linearise_at.mean if linearise_at is not None else rv.mean
        jacob = self.disc_model.jacobfun(start, compute_jacobian_at)
        mpred = self.disc_model.dynamicsfun(start, rv.mean)
        crosscov = rv.cov @ jacob.T
        cpred = jacob @ crosscov + diffmat
        return pnrv.Normal(mpred, cpred), {"crosscov": crosscov}

    @property
    def dimension(self):
        raise NotImplementedError

    @classmethod
    def from_ode(cls, ode, prior, evlvar, ek0_or_ek1=0):
        spatialdim = prior.spatialdim
        h0 = prior.proj2coord(coord=0)
        h1 = prior.proj2coord(coord=1)

        def dyna(t, x):
            return h1 @ x - ode.rhs(t, h0 @ x)

        def diff(t):
            return evlvar * np.eye(spatialdim)

        def jaco_ek1(t, x):
            return h1 - ode.jacobian(t, h0 @ x) @ h0

        def jaco_ek0(t, x):
            return h1

        if ek0_or_ek1 == 0:
            jaco = jaco_ek0
        elif ek0_or_ek1 == 1:
            jaco = jaco_ek1
        else:
            raise TypeError("ek0_or_ek1 must be 0 or 1, resp.")

        discrete_model = statespace.DiscreteGaussian(dyna, diff, jaco)
        return cls(discrete_model)



"""Finite-dimensional linear operators."""

import warnings

import numpy as np
import scipy.sparse.linalg
import scipy.sparse.linalg.interface


class LinearOperator(scipy.sparse.linalg.LinearOperator):
    """Composite base class for finite-dimensional linear operators.

    This class provides a way to define finite-dimensional linear operators without
    explicitly constructing a matrix representation. Instead it suffices to define a
    matrix-vector product and a shape attribute. This avoids unnecessary memory usage
    and can often be more convenient to derive.

    :class:`LinearOperator` instances can be multiplied, added and exponentiated. This
    happens lazily: the result of these operations is a new, composite
    :class:`LinearOperator`, that defers linear operations to the original operators and
    combines the results.

    To construct a concrete class:`LinearOperator`, either pass appropriate callables to
    the constructor of this class, or subclass it.

    A subclass must implement either one of the methods ``_matvec`` and ``_matmat``, and
    the attributes/properties ``shape`` (pair of integers) and ``dtype`` (may be
    ``None``). It may call the ``__init__`` on this class to have these attributes
    validated. Implementing ``_matvec`` automatically implements ``_matmat`` (using a
    naive algorithm) and vice-versa.

    Optionally, a subclass may implement ``_rmatvec`` or ``_adjoint`` to implement the
    Hermitian adjoint (conjugate transpose). As with ``_matvec`` and ``_matmat``,
    implementing either ``_rmatvec`` or ``_adjoint`` implements the other automatically.
    Implementing ``_adjoint`` is preferable; ``_rmatvec`` is mostly there for backwards
    compatibility.

    Parameters
    ----------
    shape : tuple
        Matrix dimensions (M, N).
    matvec : callable f(v)
        Returns :math:`A v`.
    rmatvec : callable f(v)
        Returns :math:`A^H v`, where :math:`A^H` is the conjugate transpose of
        :math:`A`.
    matmat : callable f(V)
        Returns :math:`AV`, where :math:`V` is a dense matrix with dimensions (N, K).
    dtype : dtype
        Data type of the operator.
    rmatmat : callable f(V)
        Returns :math:`A^H V`, where :math:`V` is a dense matrix with dimensions (M, K).

    See Also
    --------
    aslinop : Transform into a LinearOperator.

    Examples
    --------
    >>> import numpy as np
    >>> from probnum.linops import LinearOperator
    >>> def mv(v):
    ...     return np.array([2 * v[0] - v[1], 3 * v[1]])
    ...
    >>> A = LinearOperator(shape=(2, 2), matvec=mv)
    >>> A
    <2x2 _CustomLinearOperator with dtype=float64>
    >>> A.matvec(np.array([1., 2.]))
    array([0., 6.])
    >>> A @ np.ones(2)
    array([1., 3.])
    """

    def __new__(cls, *args, **kwargs):
        if cls is LinearOperator:
            # _CustomLinearOperator factory
            return super().__new__(_CustomLinearOperator)
        else:
            obj = super().__new__(cls)

            if (
                type(obj)._matvec == scipy.sparse.linalg.LinearOperator._matvec
                and type(obj)._matmat == scipy.sparse.linalg.LinearOperator._matmat
            ):
                warnings.warn(
                    "LinearOperator subclass should implement"
                    " at least one of _matvec and _matmat.",
                    category=RuntimeWarning,
                    stacklevel=2,
                )

            return obj

    # Overload arithmetic operators to give access to newly implemented functions (e.g.
    # todense())
    def __rmul__(self, x):
        if np.isscalar(x):
            return _ScaledLinearOperator(self, x)
        else:
            return NotImplemented

    def __pow__(self, p):
        if np.isscalar(p):
            return _PowerLinearOperator(self, p)
        else:
            return NotImplemented

    def __add__(self, x):
        if isinstance(x, LinearOperator):
            return _SumLinearOperator(self, x)
        else:
            return NotImplemented

    def __neg__(self):
        return _ScaledLinearOperator(self, -1)

    # The below methods are overloaded to allow dot products with random variables
    def dot(self, x):
        """Matrix-matrix or matrix-vector multiplication.

        Parameters
        ----------
        x : array_like
            1-d or 2-d array, representing a vector or matrix.

        Returns
        -------
        Ax : array
            1-d or 2-d array (depending on the shape of x) that represents
            the result of applying this linear operator on x.
        """
        if isinstance(x, LinearOperator):
            return _ProductLinearOperator(self, x)
        elif np.isscalar(x):
            return _ScaledLinearOperator(self, x)
        else:
            if len(x.shape) == 1 or len(x.shape) == 2 and x.shape[1] == 1:
                return self.matvec(x)
            elif len(x.shape) == 2:
                return self.matmat(x)
            else:
                raise ValueError(
                    "Expected 1-d or 2-d array, matrix or random variable, got %r." % x
                )

    def matvec(self, x):
        """Matrix-vector multiplication. Performs the operation y=A*x where A is an MxN
        linear operator and x is a 1-d array or random variable.

        Parameters
        ----------
        x : {matrix, ndarray, RandomVariable}
            An array or RandomVariable with shape (N,) or (N,1).
        Returns
        -------
        y : {matrix, ndarray}
            A matrix or ndarray or RandomVariable with shape (M,) or (M,1) depending
            on the type and shape of the x argument.
        Notes
        -----
        This matvec wraps the user-specified matvec routine or overridden
        _matvec method to ensure that y has the correct shape and type.
        """
        M, N = self.shape

        if x.shape != (N,) and x.shape != (N, 1):
            raise ValueError("Dimension mismatch.")

        y = self._matvec(x)

        if isinstance(x, np.matrix):
            y = scipy.sparse.sputils.asmatrix(y)

        if isinstance(x, (np.matrix, np.ndarray)):
            if x.ndim == 1:
                y = y.reshape(M)
            elif x.ndim == 2:
                y = y.reshape(M, 1)
            else:
                raise ValueError("Invalid shape returned by user-defined matvec().")
        # TODO: can be shortened once RandomVariable implements a reshape method
        elif y.shape[0] != M:
            raise ValueError("Invalid shape returned by user-defined matvec().")

        return y

    def transpose(self):
        """Transpose this linear operator.

        Can be abbreviated self.T instead of self.transpose().
        """
        return self._transpose()

    T = property(transpose)

    def _transpose(self):
        """Default implementation of _transpose; defers to rmatvec + conj."""
        return _TransposedLinearOperator(self)

    def todense(self):
        """Dense matrix representation of the linear operator.

        This method can be computationally very costly depending on the shape of the
        linear operator. Use with caution.

        Returns
        -------
        matrix : np.ndarray
            Matrix representation of the linear operator.
        """
        return self.matmat(np.eye(self.shape[1], dtype=self.dtype))

    def inv(self):
        """Inverse of the linear operator."""
        raise NotImplementedError

    # TODO: implement operations (eigs, cond, det, logabsdet, trace, ...)
    def rank(self):
        """Rank of the linear operator."""
        raise NotImplementedError

    def eigvals(self):
        """Eigenvalue spectrum of the linear operator."""
        raise NotImplementedError

    def cond(self, p=None):
        """Compute the condition number of the linear operator.

        The condition number of the linear operator with respect to the ``p`` norm. It
        measures how much the solution :math:`x` of the linear system :math:`Ax=b`
        changes with respect to small changes in :math:`b`.

        Parameters
        ----------
        p : {None, 1, , 2, , inf, 'fro'}, optional
            Order of the norm:

            =======  ============================
            p        norm for matrices
            =======  ============================
            None     2-norm, computed directly via singular value decomposition
            'fro'    Frobenius norm
            np.inf   max(sum(abs(x), axis=1))
            1        max(sum(abs(x), axis=0))
            2        2-norm (largest sing. value)
            =======  ============================

        Returns
        -------
        cond : {float, inf}
            The condition number of the linear operator. May be infinite.
        """
        raise NotImplementedError

    def det(self):
        """Determinant of the linear operator."""
        raise NotImplementedError

    def logabsdet(self):
        """Log absolute determinant of the linear operator."""
        raise NotImplementedError

    def trace(self):
        """Trace of the linear operator.

        Computes the trace of a square linear operator :math:`\\text{tr}(A) =
        \\sum_{i-1}^n A_ii`.

        Returns
        -------
        trace : float
            Trace of the linear operator.

        Raises
        ------
        ValueError : If :meth:`trace` is called on a non-square matrix.
        """
        if self.shape[0] != self.shape[1]:
            raise ValueError("The trace is only defined for square linear operators.")
        else:
            _identity = np.eye(self.shape[0])
            trace = 0.0
            for i in range(self.shape[0]):
                trace += np.squeeze(
                    _identity[np.newaxis, i, :]
                    @ self.matvec(_identity[i, :, np.newaxis])
                )
            return trace


class _CustomLinearOperator(
    scipy.sparse.linalg.interface._CustomLinearOperator, LinearOperator
):
    """Linear operator defined in terms of user-specified operations."""

    def __init__(
        self, shape, matvec, rmatvec=None, matmat=None, rmatmat=None, dtype=None
    ):
        super().__init__(
            shape=shape,
            matvec=matvec,
            rmatvec=rmatvec,
            matmat=matmat,
            rmatmat=rmatmat,
            dtype=dtype,
        )


# TODO: inheritance from _TransposedLinearOperator causes dependency on scipy>=1.4,
# maybe implement our own instead?
class _TransposedLinearOperator(
    scipy.sparse.linalg.interface._TransposedLinearOperator, LinearOperator
):
    """Transposition of a linear operator."""

    def __init__(self, A):
        self.A = A
        super().__init__(A=A)

    def todense(self):
        return self.A.todense().T

    def inv(self):
        return self.A.inv().T


class _SumLinearOperator(
    scipy.sparse.linalg.interface._SumLinearOperator, LinearOperator
):
    """Sum of two linear operators."""

    def __init__(self, A, B):
        self.A = A
        self.B = B
        super().__init__(A=A, B=B)

    def todense(self):
        return self.A.todense() + self.B.todense()

    def inv(self):
        return self.A.inv() + self.B.inv()

    def trace(self):
        return self.A.trace() + self.B.trace()


class _ProductLinearOperator(
    scipy.sparse.linalg.interface._ProductLinearOperator, LinearOperator
):
    """(Operator) Product of two linear operators."""

    def __init__(self, A, B):
        self.A = A
        self.B = B
        super().__init__(A=A, B=B)

    def todense(self):
        return self.A.todense() @ self.B.todense()


class _ScaledLinearOperator(
    scipy.sparse.linalg.interface._ScaledLinearOperator, LinearOperator
):
    """Linear operator scaled with a scalar."""

    def __init__(self, A, alpha):
        super().__init__(A=A, alpha=alpha)

    def todense(self):
        A, alpha = self.args
        return alpha * A.todense()

    def inv(self):
        A, alpha = self.args
        return _ScaledLinearOperator(A.inv(), 1 / alpha)

    def trace(self):
        A, alpha = self.args
        return alpha * A.trace()


class _PowerLinearOperator(
    scipy.sparse.linalg.interface._PowerLinearOperator, LinearOperator
):
    """Linear operator raised to a non-negative integer power."""

    def __init__(self, A, p):
        super().__init__(A=A, p=p)


class Diagonal(LinearOperator):
    """A linear operator representing the diagonal from another linear operator.

    Parameters
    ----------
    Op : LinearOperator
        Linear operator of which to represent the diagonal.
    """

    # TODO: should this be an operator itself or a function of a LinearOperator?
    #   - a function allows subclasses (e.g. MatrixMult) to implement more efficient
    # versions than n products e_i A e_i
    def __init__(self, Op):
        # pylint: disable=super-init-not-called
        raise NotImplementedError


class ScalarMult(LinearOperator):
    """A linear operator representing scalar multiplication.

    Parameters
    ----------
    shape : tuple
        Matrix dimensions (M, N).
    scalar : float
        Scalar to multiply by.
    """

    def __init__(self, shape, scalar):
        self.scalar = scalar
        super().__init__(shape=shape, dtype=float)

    def _matvec(self, x):
        return self.scalar * x

    def _matmat(self, X):
        return self.scalar * X

    def todense(self):
        return np.eye(self.shape[0]) * self.scalar

    def inv(self):
        return ScalarMult(shape=self.shape, scalar=1 / self.scalar)

    # Properties
    def rank(self):
        return np.minimum(self.shape[0], self.shape[1])

    def eigvals(self):
        return np.ones(self.shape[0]) * self.scalar

    def cond(self, p=None):
        return 1

    def det(self):
        return self.scalar ** self.shape[0]

    def logabsdet(self):
        return np.log(np.abs(self.scalar))

    def trace(self):
        return self.scalar * self.shape[0]


class Identity(ScalarMult):
    """The identity operator.

    Parameters
    ----------
    shape : int or tuple
        Shape of the identity operator.
    """

    def __init__(self, shape):
        # Check shape
        if np.isscalar(shape):
            _shape = (shape, shape)
        elif shape[0] != shape[1]:
            raise ValueError("The identity operator must be square.")
        else:
            _shape = shape
        # Initiator of super class
        super().__init__(shape=_shape, scalar=1.0)

    def todense(self):
        return np.eye(self.shape[0])

    def inv(self):
        return self

    # Properties
    def rank(self):
        return self.shape[0]

    def eigvals(self):
        return np.ones(self.shape[0])

    def cond(self, p=None):
        return 1

    def det(self):
        return 1.0

    def logabsdet(self):
        return 0.0

    def trace(self):
        return self.shape[0]


class MatrixMult(scipy.sparse.linalg.interface.MatrixLinearOperator, LinearOperator):
    """A linear operator defined via a matrix.

    Parameters
    ----------
    A : array-like or scipy.sparse.spmatrix
        The explicit matrix.
    """

    def __init__(self, A):
        super().__init__(A=A)

    def _matvec(self, x):
        return self.A @ x  # Needed to call __matmul__ instead of np.dot or np.matmul

    def _matmat(self, X):
        return self.A @ X

    def todense(self):
        if isinstance(self.A, scipy.sparse.spmatrix):
            return self.A.todense()
        else:
            return np.asarray(self.A)

    def inv(self):
        if isinstance(self.A, scipy.sparse.spmatrix):
            invmat = scipy.sparse.linalg.inv(self.A)
        else:
            invmat = np.linalg.inv(self.A)
        return MatrixMult(invmat)

    # Arithmetic operations
    # TODO: perform arithmetic operations between MatrixMult operators explicitly

    # Properties
    def rank(self):
        return np.linalg.matrix_rank(self.A)

    def eigvals(self):
        return np.linalg.eigvals(self.A)

    def cond(self, p=None):
        return np.linalg.cond(self.A, p=p)

    def det(self):
        return np.linalg.det(self.A)

    def logabsdet(self):
        _sign, logdet = np.linalg.slogdet(self.A)
        return logdet

    def trace(self):
        if self.shape[0] != self.shape[1]:
            raise ValueError("The trace is only defined for square linear operators.")
        else:
            return np.trace(self.A)



import numpy as np

import probnum.filtsmooth as pnfs
from probnum.diffeq import odesolver
from probnum.diffeq.odesolution import ODESolution
from probnum.random_variables import Normal


class GaussianIVPFilter(odesolver.ODESolver):
    """ODE solver that behaves like a Gaussian filter.

    This is based on continuous-discrete Gaussian filtering.

    Note: this is specific for IVPs and does not apply without
    further considerations to, e.g., BVPs.

    Parameters
    ----------
    gaussfilt : gaussianfilter.GaussianFilter
        e.g. the return value of ivp_to_ukf(), ivp_to_ekf1().

    Notes
    -----
    - gaussfilt.dynamicmodel contains the prior,
    - gaussfilt.measurementmodel contains the information about the ODE right hand side function,
    - gaussfilt.initialdistribution contains the information about the initial values.
    """

    def __init__(self, ivp, gaussfilt, with_smoothing):
        if not isinstance(gaussfilt.dynamics_model, pnfs.statespace.Integrator):
            raise ValueError(
                "Please initialise a Gaussian filter with an Integrator (see filtsmooth.statespace)"
            )
        self.gfilt = gaussfilt
        self.sigma_squared_mle = 1.0
        self.with_smoothing = with_smoothing
        super().__init__(ivp=ivp, order=gaussfilt.dynamics_model.ordint)

    def initialise(self):
        return self.ivp.t0, self.gfilt.initrv

    def step(self, t, t_new, current_rv):
        """Gaussian IVP filter step as nonlinear Kalman filtering with zero data."""

        # Read the diffusion matrix; required for calibration / error estimation
        discrete_dynamics = self.gfilt.dynamics_model.discretise(t_new - t)
        diffmat = discrete_dynamics.diffmat

        # 1. Predict
        pred_rv, _ = self.gfilt.predict(t, t_new, current_rv)

        # 2. Measure
        meas_rv, info = self.gfilt.measure(t_new, pred_rv)

        # 3. Estimate the diffusion (sigma squared)
        self.sigma_squared_mle = self._estimate_diffusion(meas_rv)
        # 3.1. Adjust the prediction covariance to include the diffusion
        pred_rv = Normal(
            pred_rv.mean, pred_rv.cov + (self.sigma_squared_mle - 1) * diffmat
        )
        # 3.2 Update the measurement covariance (measure again)
        meas_rv, info = self.gfilt.measure(t_new, pred_rv)

        # 4. Update
        zero_data = 0.0
        filt_rv = self.gfilt.condition_state_on_measurement(
            pred_rv, meas_rv, zero_data, info["crosscov"]
        )

        # 5. Error estimate
        local_errors = self._estimate_local_error(
            pred_rv, t_new, self.sigma_squared_mle * diffmat
        )
        err = np.linalg.norm(local_errors)

        return filt_rv, err

    def postprocess(self, times, rvs):
        """Rescale covariances with sigma square estimate, (if specified) smooth the
        estimate, return ODESolution."""
        if False:  # pylint: disable=using-constant-test
            # will become useful again for time-fixed diffusion models
            rvs = self._rescale(rvs)
        odesol = super().postprocess(times, rvs)
        if self.with_smoothing is True:
            odesol = self._odesmooth(ode_solution=odesol)
        return odesol

    def _rescale(self, rvs):
        """Rescales covariances according to estimate sigma squared value."""
        rvs = [Normal(rv.mean, self.sigma_squared_mle * rv.cov) for rv in rvs]
        return rvs

    def _odesmooth(self, ode_solution, **kwargs):
        """Smooth out the ODE-Filter output.

        Be careful about the preconditioning: the GaussFiltSmooth object
        only knows the state space with changed coordinates!

        Parameters
        ----------
        filter_solution: ODESolution

        Returns
        -------
        smoothed_solution: ODESolution
        """
        ivp_filter_posterior = ode_solution._kalman_posterior
        ivp_smoother_posterior = self.gfilt.smooth(ivp_filter_posterior)

        smoothed_solution = ODESolution(
            times=ivp_smoother_posterior.locations,
            rvs=ivp_smoother_posterior.state_rvs,
            solver=ode_solution._solver,
        )

        return smoothed_solution

    def _estimate_local_error(self, pred_rv, t_new, calibrated_diffmat, **kwargs):
        """Estimate the local errors.

        This corresponds to the approach in [1], implemented such that it is compatible
        with the EKF1 and UKF.

        References
        ----------
        .. [1] Schober, M., Särkkä, S. and Hennig, P..
            A probabilistic model for the numerical solution of initial
            value problems.
            Statistics and Computing, 2019.
        """
        local_pred_rv = Normal(pred_rv.mean, calibrated_diffmat)
        local_meas_rv, _ = self.gfilt.measure(t_new, local_pred_rv)
        error = local_meas_rv.cov.diagonal()
        return np.sqrt(error)

    def _estimate_diffusion(self, meas_rv):
        """Estimate the dynamic diffusion parameter sigma_squared.

        This corresponds to the approach in [1], implemented such that it is compatible
        with the EKF1 and UKF.

        References
        ----------
        .. [1] Schober, M., Särkkä, S. and Hennig, P..
            A probabilistic model for the numerical solution of initial
            value problems.
            Statistics and Computing, 2019.
        """
        std_like = np.linalg.cholesky(meas_rv.cov)
        whitened_res = np.linalg.solve(std_like, meas_rv.mean)
        ssq = whitened_res @ whitened_res / meas_rv.size
        return ssq

    @property
    def prior(self):
        return self.gfilt.dynamics_model



from .ivp import IVP
from .ivp_examples import (
    fitzhughnagumo,
    logistic,
    lotkavolterra,
    rigidbody,
    seir,
    threebody,
    vanderpol,
)
from .ode import ODE

__all__ = [
    "ODE",
    "IVP",
    "logistic",
    "fitzhughnagumo",
    "seir",
    "rigidbody",
    "vanderpol",
    "threebody",
    "lotkavolterra",
]



"""Warped Bayesian Quadrature."""

from ._bayesian_quadrature import BayesianQuadrature


class WarpedBayesianQuadrature(BayesianQuadrature):
    """Warped Sequential Active Bayesian Integration (WSABI)."""

    def integrate(self, fun, domain, measure, nevals):
        """Integrate the function ``fun``.

        Parameters
        ----------
        fun :
            Function to be integrated.
        domain :
            Domain to integrate over.
        measure :
            Measure to integrate against.
        nevals :
            Number of function evaluations.

        Returns
        -------
        """
        raise NotImplementedError



"""Convenience functions for Gaussian filtering and smoothing.

We support the following methods:
    - ekf0: Extended Kalman filtering based on a zero-th order Taylor
        approximation [1]_, [2]_, [3]_. Also known as "PFOS".
    - ekf1: Extended Kalman filtering [3]_.
    - ukf: Unscented Kalman filtering [3]_.
    - eks0: Extended Kalman smoothing based on a zero-th order Taylor
        approximation [4]_.
    - eks1: Extended Kalman smoothing [4]_.
    - uks: Unscented Kalman smoothing.

References
----------
.. [1] https://arxiv.org/pdf/1610.05261.pdf
.. [2] https://arxiv.org/abs/1807.09737
.. [3] https://arxiv.org/abs/1810.03440
.. [4] https://arxiv.org/pdf/2004.00623.pdf
"""

import numpy as np

import probnum.filtsmooth as pnfs
from probnum.diffeq import steprule
from probnum.diffeq.odefiltsmooth import ivp2filter
from probnum.diffeq.odefiltsmooth.ivpfiltsmooth import GaussianIVPFilter


def probsolve_ivp(
    ivp,
    method="ekf0",
    which_prior="ibm1",
    atol=None,
    rtol=None,
    step=None,
    firststep=None,
    **kwargs
):
    """Solve initial value problem with Gaussian filtering and smoothing.

    Numerically computes a Gauss-Markov process which solves numerically
    the initial value problem (IVP) based on a system of first order
    ordinary differential equations (ODEs)

    .. math:: \\dot x(t) = f(t, x(t)), \\quad x(t_0) = x_0,
        \\quad t \\in [t_0, T]

    by regarding it as a (nonlinear) Gaussian filtering (and smoothing)
    problem [3]_. For some configurations it recovers certain multistep
    methods [1]_.
    Convergence rates of filtering [2]_ and smoothing [4]_ are
    comparable to those of methods of Runge-Kutta type.


    This function turns a prior-string into an :class:`ODEPrior`, a
    method-string into a filter/smoother of class :class:`GaussFiltSmooth`, creates a
    :class:`GaussianIVPFilter` object and calls the :meth:`solve()` method. For
    advanced usage we recommend to do this process manually which
    enables advanced methods of tuning the algorithm.

    This function supports the methods:
    extended Kalman filtering based on a zero-th order Taylor
    approximation (EKF0),
    extended Kalman filtering (EKF1),
    unscented Kalman filtering (UKF),
    extended Kalman smoothing based on a zero-th order Taylor
    approximation (EKS0),
    extended Kalman smoothing (EKS1), and
    unscented Kalman smoothing (UKS).

    Arguments
    ---------
    ivp : IVP
        Initial value problem to be solved.
    step : float
        Step size :math:`h` of the solver. This defines the
        discretisation mesh as each proposed step is equal to :math:`h`
        and all proposed steps are accepted.
        Only one of out of ``step`` and ``tol`` is set.
    tol : float
        Tolerance :math:`\\varepsilon` of the adaptive step scheme.
        We implement the scheme proposed by Schober et al., accepting a
        step if the absolute as well as the relative error estimate are
        smaller than the tolerance,
        :math:`\\max\\{e, e / |y|\\} \\leq \\varepsilon`.
        Only one of out of ``step`` and ``tol`` is set.
    which_prior : str, optional
        Which prior is to be used. Default is an IBM(1), further support
        for IBM(:math:`q`), IOUP(:math:`q`), Matern(:math:`q+1/2`),
        :math:`q\\in\\{1, 2, 3, 4\\}` is provided. The available
        options are

        ======================  ========================================
         IBM(:math:`q`)         ``'ibm1'``, ``'ibm2'``, ``'ibm3'``,
                                ``'ibm4'``
         IOUP(:math:`q`)        ``'ioup1'``, ``'ioup2'``, ``'ioup3'``,
                                ``'ioup4'``
         Matern(:math:`q+0.5`)  ``'matern32'``, ``'matern52'``,
                                ``'matern72'``, ``'matern92'``
        ======================  ========================================

        The type of prior relates to prior assumptions about the
        derivative of the solution. The IBM(:math:`q`) prior leads to a
        :math:`q`-th order method that is recommended if little to no
        prior information about the solution is available. On the other
        hand, if the :math:`q`-th derivative is expected to regress to
        zero, an IOUP(:math:`q`) prior might be suitable.
    method : str, optional
        Which method is to be used. Default is ``ekf0`` which is the
        method proposed by Schober et al.. The available
        options are

        ================================================  ==============
         Extended Kalman filtering/smoothing (0th order)  ``'ekf0'``,
                                                          ``'eks0'``
         Extended Kalman filtering/smoothing (1st order)  ``'ekf1'``,
                                                          ``'eks1'``
         Unscented Kalman filtering/smoothing             ``'ukf'``,
                                                          ``'uks'``
        ================================================  ==============

        First order extended Kalman filtering and smoothing methods
        require Jacobians of the RHS-vector field of the IVP. The
        uncertainty estimates as returned by EKF1/S1 and UKF/S appear to
        be more reliable than those of EKF0/S0. The latter is more
        stable when it comes to very small steps.

    firststep : float, optional
        First suggested step :math:`h_0` for adaptive step size scheme.
        Default is None which lets the solver start with the suggestion
        :math:`h_0 = T - t_0`. For low accuracy it might be more
        efficient to start out with smaller :math:`h_0` so that the
        first acceptance occurs earlier.

    Returns
    -------
    solution : ODESolution
        Solution of the ODE problem.

        Contains fields:

        t : :obj:`np.ndarray`, shape=(N,)
            Mesh used by the solver to compute the solution.
            It includes the initial time :math:`t_0` but not necessarily the
            final time :math:`T`.
        y : :obj:`list` of :obj:`RandomVariable`, length=N
            Discrete-time solution at times :math:`t_1, ..., t_N`,
            as a list of random variables.
            The means and covariances can be accessed with ``solution.y.mean``
            and ``solution.y.cov``.

    See Also
    --------
    GaussianIVPFilter : Solve IVPs with Gaussian filtering and smoothing
    ODESolution : Solution of ODE problems

    References
    ----------
    .. [1] Schober, M., Särkkä, S. and Hennig, P..
        A probabilistic model for the numerical solution of initial
        value problems.
        Statistics and Computing, 2019.
    .. [2] Kersting, H., Sullivan, T.J., and Hennig, P..
        Convergence rates of Gaussian ODE filters.
        2019.
    .. [3] Tronarp, F., Kersting, H., Särkkä, S., and Hennig, P..
        Probabilistic solutions to ordinary differential equations as
        non-linear Bayesian filtering: a new perspective.
        Statistics and Computing, 2019.
    .. [4] Tronarp, F., Särkkä, S., and Hennig, P..
        Bayesian ODE solvers: the maximum a posteriori estimate.
        2019.


    Examples
    --------
    >>> from probnum.diffeq import logistic, probsolve_ivp
    >>> from probnum import random_variables as rvs
    >>> import numpy as np
    >>> initrv = rvs.Constant(0.15)
    >>> ivp = logistic(timespan=[0., 1.5], initrv=initrv, params=(4, 1))
    >>> solution = probsolve_ivp(ivp, method="ekf0", step=0.1)
    >>> print(np.round(solution.y.mean, 2))
    [[0.15]
     [0.21]
     [0.28]
     [0.36]
     [0.46]
     [0.56]
     [0.65]
     [0.74]
     [0.81]
     [0.86]
     [0.9 ]
     [0.93]
     [0.95]
     [0.97]
     [0.98]
     [0.98]]

    >>> initrv = rvs.Constant(0.15)
    >>> ivp = logistic(timespan=[0., 1.5], initrv=initrv, params=(4, 1))
    >>> solution = probsolve_ivp(ivp, method="eks1", which_prior="ioup3", step=0.1)
    >>> print(np.round(solution.y.mean, 2))
    [[0.15]
     [0.21]
     [0.28]
     [0.37]
     [0.47]
     [0.57]
     [0.66]
     [0.74]
     [0.81]
     [0.87]
     [0.91]
     [0.93]
     [0.96]
     [0.97]
     [0.98]
     [0.99]]
    """
    stprl = _create_steprule(atol, rtol, step, firststep, ivp)
    prior = _string2prior(ivp, which_prior, **kwargs)
    gfilt = _create_filter(ivp, prior, method, **kwargs)
    with_smoothing = method[-2] == "s" or method[-1] == "s"
    solver = GaussianIVPFilter(ivp, gfilt, with_smoothing=with_smoothing)
    solution = solver.solve(steprule=stprl)
    return solution


def _create_filter(ivp, prior, method, **kwargs):
    """Create the solver object that is used."""
    if method not in ["ekf0", "ekf1", "ukf", "eks0", "eks1", "uks"]:
        raise ValueError("Method not supported.")
    gfilt = _string2filter(ivp, prior, method, **kwargs)
    return gfilt


def _create_steprule(atol, rtol, step, firststep, ivp):
    _check_step_tol(step, atol, rtol)

    if step is not None:
        stprl = steprule.ConstantSteps(step)
    else:
        if firststep is None:
            # lazy version of Hairer, Wanner, Norsett, p. 169
            norm_y0 = np.linalg.norm(ivp.initrv.mean)
            norm_dy0 = np.linalg.norm(ivp(ivp.t0, ivp.initrv.mean))
            firststep = 0.01 * norm_y0 / norm_dy0
        stprl = steprule.AdaptiveSteps(firststep=firststep, atol=atol, rtol=rtol)
    return stprl


def _check_step_tol(step, atol, rtol):
    both_none = atol is None and rtol is None and step is None
    both_not_none = (atol is not None and rtol is not None) and step is not None
    if both_none or both_not_none:
        errormsg = "Please specify either a tolerance or a step size."
        raise ValueError(errormsg)
    atol_not_rtol = atol is not None and rtol is None
    rtol_not_atol = rtol is not None and atol is None
    if atol_not_rtol or rtol_not_atol:
        errormsg = "Please specify either both atol and rtol, or neither."
        raise ValueError(errormsg)


def _string2prior(ivp, which_prior, **kwargs):

    ibm_family = ["ibm1", "ibm2", "ibm3", "ibm4"]
    ioup_family = ["ioup1", "ioup2", "ioup3", "ioup4"]
    matern_family = ["matern32", "matern52", "matern72", "matern92"]
    if which_prior in ibm_family:
        return _string2ibm(ivp, which_prior, **kwargs)
    elif which_prior in ioup_family:
        return _string2ioup(ivp, which_prior, **kwargs)
    elif which_prior in matern_family:
        return _string2matern(ivp, which_prior, **kwargs)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2ibm(ivp, which_prior, **kwargs):

    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if which_prior == "ibm1":
        return pnfs.statespace.IBM(1, ivp.dimension, diffconst)
    elif which_prior == "ibm2":
        return pnfs.statespace.IBM(2, ivp.dimension, diffconst)
    elif which_prior == "ibm3":
        return pnfs.statespace.IBM(3, ivp.dimension, diffconst)
    elif which_prior == "ibm4":
        return pnfs.statespace.IBM(4, ivp.dimension, diffconst)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2ioup(ivp, which_prior, **kwargs):

    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if "driftspeed" in kwargs.keys():
        driftspeed = kwargs["driftspeed"]
    else:
        driftspeed = 1.0
    if which_prior == "ioup1":
        return pnfs.statespace.IOUP(1, ivp.dimension, driftspeed, diffconst)
    elif which_prior == "ioup2":
        return pnfs.statespace.IOUP(2, ivp.dimension, driftspeed, diffconst)
    elif which_prior == "ioup3":
        return pnfs.statespace.IOUP(3, ivp.dimension, driftspeed, diffconst)
    elif which_prior == "ioup4":
        return pnfs.statespace.IOUP(4, ivp.dimension, driftspeed, diffconst)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2matern(ivp, which_prior, **kwargs):

    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if "lengthscale" in kwargs.keys():
        lengthscale = kwargs["lengthscale"]
    else:
        lengthscale = 1.0
    if which_prior == "matern32":
        return pnfs.statespace.Matern(1, ivp.dimension, lengthscale, diffconst)
    elif which_prior == "matern52":
        return pnfs.statespace.Matern(2, ivp.dimension, lengthscale, diffconst)
    elif which_prior == "matern72":
        return pnfs.statespace.Matern(3, ivp.dimension, lengthscale, diffconst)
    elif which_prior == "matern92":
        return pnfs.statespace.Matern(4, ivp.dimension, lengthscale, diffconst)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2filter(_ivp, _prior, _method, **kwargs):

    if "evlvar" in kwargs.keys():
        evlvar = kwargs["evlvar"]
    else:
        evlvar = 0.0
    if _method in ("ekf0", "eks0"):
        return ivp2filter.ivp2ekf0(_ivp, _prior, evlvar)
    elif _method in ("ekf1", "eks1"):
        return ivp2filter.ivp2ekf1(_ivp, _prior, evlvar)
    elif _method in ("ukf", "uks"):
        return ivp2filter.ivp2ukf(_ivp, _prior, evlvar)
    else:
        raise ValueError("Type of filter not supported.")



"""Bayesian Filtering and Smoothing."""

from .bayesfiltsmooth import BayesFiltSmooth
from .filtsmoothposterior import FiltSmoothPosterior
from .gaussfiltsmooth import (
    ContinuousEKFComponent,
    ContinuousUKFComponent,
    DiscreteEKFComponent,
    DiscreteUKFComponent,
    FixedPointStopping,
    IteratedKalman,
    Kalman,
    KalmanPosterior,
    StoppingCriterion,
    UnscentedTransform,
)

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "Kalman",
    "ContinuousEKFComponent",
    "DiscreteEKFComponent",
    "ContinuousUKFComponent",
    "DiscreteUKFComponent",
    "UnscentedTransform",
    "FiltSmoothPosterior",
    "KalmanPosterior",
    "IteratedKalman",
    "StoppingCriterion",
    "FixedPointStopping",
]



"""Implementation of Bayesian Quadrature."""


class BayesianQuadrature:
    """Bayesian quadrature.

    Bayesian quadrature methods build a model for the integrand via function
    evaluations and return a belief over the value of the integral on a given
    domain with respect to the specified measure.

    Parameters
    ----------
    fun0
        Stochastic process modelling function to be integrated.
    """

    def __init__(self, fun0):
        self.fun0 = fun0

    def integrate(self, fun, domain, measure, nevals):
        """Integrate the function ``fun``.

        Parameters
        ----------
        fun :
            Function to be integrated.
        domain :
            Domain to integrate over.
        measure :
            Measure to integrate against.
        nevals :
            Number of function evaluations.

        Returns
        -------
        """
        raise NotImplementedError



"""Bayesian Quadrature.

This module provides routines to integrate functions through Bayesian
quadrature, meaning a model over the integrand is constructed in order
to actively select evaluation points of the integrand to estimate the
value of the integral. Bayesian quadrature methods return a random
variable, specifying the belief about the true value of the integral.
"""

from .bq_methods import BayesianQuadrature, WarpedBayesianQuadrature


def bayesquad(fun, fun0, domain, measure, nevals=None, method="vanilla"):
    """N-dimensional Bayesian quadrature.

    Parameters
    ----------
    fun :
        Function to be integrated.
    fun0 :
        Stochastic process modelling the function to be integrated.
    domain :
        Domain of integration.
    measure :
        Measure to integrate against.
    nevals :
        Number of function evaluations.
    method :
        Type of Bayesian quadrature to use. The available options are

        ====================  ===========
         vanilla              ``vanilla``
         WSABI                ``wsabi``
        ====================  ===========

    Returns
    -------
    integral :
        The integral of ``func`` on the domain.
    fun0 :
        Stochastic process modelling the function to be integrated after ``neval``
        observations.
    info :
        Information on the performance of the method.

    References
    ----------
    """

    # Choose Method
    bqmethod = None
    if method == "vanilla":
        bqmethod = BayesianQuadrature(fun0=fun0)
    elif method == "wsabi":
        bqmethod = WarpedBayesianQuadrature(fun0=fun0)

    # Integrate
    integral, fun0, info = bqmethod.integrate(
        fun=fun, nevals=nevals, domain=domain, measure=measure
    )

    return integral, fun0, info



"""Definitions of problems currently solved by probabilistic numerical methods."""

import dataclasses
import typing

import numpy as np
import scipy.sparse

import probnum.filtsmooth as pnfs
import probnum.linops as pnlo
import probnum.random_variables as pnrv
import probnum.type as pntp


@dataclasses.dataclass
class RegressionProblem:
    r"""Regression problem.

    Fit a stochastic process to data, given a likelihood (realised by a :obj:`DiscreteGaussian` transition).
    Solved by filters and smoothers in :mod:`probnum.filtsmooth`.

    Parameters
    ----------
    observations
        Observations of the latent process.
    locations
        Grid-points on which the observations were taken.
    likelihood
        Likelihood of the observations; that is, relation between the latent process and the observed values.
        Encodes for example noise.
    solution
        Closed form, analytic solution to the problem. Used for testing and benchmarking.

    Examples
    --------
    >>> obs = [11.4123, -15.5123]
    >>> loc = [0.1, 0.2]
    >>> rp = RegressionProblem(observations=obs, locations=loc)
    >>> rp
    RegressionProblem(observations=[11.4123, -15.5123], locations=[0.1, 0.2], likelihood=None, solution=None)
    >>> rp.observations
    [11.4123, -15.5123]
    """

    observations: np.ndarray
    locations: np.ndarray

    # Optional, because it should be specifiable without explicit likelihood info.
    # 'DiscreteGaussian' is currently in 'statespace', but can be used to define general
    # Likelihood functions; see #282
    likelihood: typing.Optional[pnfs.statespace.DiscreteGaussian] = None

    # For testing and benchmarking
    solution: typing.Optional[
        typing.Callable[[np.ndarray], typing.Union[float, np.ndarray]]
    ] = None


@dataclasses.dataclass
class InitialValueProblem:
    r"""First order ODE initial value problem.

    Compute a function :math:`y=y(t)` that solves

    .. math::
        \dot y(t) = f(t, y(t)), \quad y(t_0) = y_0

    on time-interval :math:`[t_0, t_\text{max}]`.
    Solved by probabilistic ODE solvers in :mod:`probnum.diffeq`.


    Parameters
    ----------
    f
        ODE vector-field.
    t0
        Initial point in time.
    tmax
        Final point in time.
    y0
        Initial value of the solution.
    df
        Jacobian of the ODE vector-field :math:`f=f(t,y)` with respect to the :math:`y` variable.
    ddf
        Hessian of the ODE vector-field :math:`f=f(t,y)` with respect to the :math:`y` variable.
    solution
        Closed form, analytic solution to the problem. Used for testing and benchmarking.

    Examples
    --------
    >>> def f(t, x):
    ...     return x*(1-x)
    >>> ivp = InitialValueProblem(f, t0=0., tmax=3., y0=0.1)
    >>> ivp.t0, ivp.tmax, ivp.y0
    (0.0, 3.0, 0.1)
    >>> np.round(ivp.f(ivp.t0, ivp.y0), 2)
    0.09
    """

    f: typing.Callable[[float, np.ndarray], np.ndarray]
    t0: float
    tmax: float
    y0: typing.Union[pntp.FloatArgType, np.ndarray]
    df: typing.Optional[typing.Callable[[float, np.ndarray], np.ndarray]] = None
    ddf: typing.Optional[typing.Callable[[float, np.ndarray], np.ndarray]] = None

    # For testing and benchmarking
    solution: typing.Optional[typing.Callable[[float, np.ndarray], np.ndarray]] = None


@dataclasses.dataclass
class LinearSystem:
    r"""Linear system of equations.

    Compute :math:`x` from :math:`Ax=b`.
    Solved by probabilistic linear solvers in :mod:`probnum.linalg`

    Parameters
    ----------
    A
        System matrix or linear operator.
    b
        Right-hand side vector or matrix.
    solution
        True solution to the problem. Used for testing and benchmarking.

    Examples
    --------
    >>> A = np.eye(3)
    >>> b = np.arange(3)
    >>> lin_sys = LinearSystem(A, b)
    >>> lin_sys
    LinearSystem(A=array([[1., 0., 0.],
           [0., 1., 0.],
           [0., 0., 1.]]), b=array([0, 1, 2]), solution=None)
    """

    A: typing.Union[
        np.ndarray,
        scipy.sparse.spmatrix,
        pnlo.LinearOperator,
        pnrv.RandomVariable,
    ]
    b: typing.Union[np.ndarray, pnrv.RandomVariable]

    # For testing and benchmarking
    solution: typing.Optional[typing.Union[np.ndarray, pnrv.RandomVariable]] = None


@dataclasses.dataclass
class QuadratureProblem:
    r"""Numerical computation of an integral.

    Compute the integral

        .. math::
            \int_\Omega f(x) \, \text{d} \mu(x)

    for a function :math:`f: \Omega \rightarrow \mathbb{R}`.
    For the time being, :math:`\mu` is the Lebesgue measure.
    Solved by quadrature rules in :mod:`probnum.quad`.

    Parameters
    ----------
    integrand
        Function to be integrated.
    lower_bd
        A number or a vector representing the lower bounds of the integrals.
    upper_bd
        A number or a vector representing the upper bounds of the integrals.
    output_dim
        Output dimension of the integrand.
    solution
        Closed form, analytic solution to the problem. Used for testing and benchmarking.

    Examples
    --------
    >>> import numpy as np
    >>> def integrand(x):
    ...     return np.linalg.norm(x)**2
    >>> lower_bd = 0.41
    >>> upper_bd = 4.32
    >>> qp1d = QuadratureProblem(integrand, lower_bd=lower_bd, upper_bd=upper_bd)
    >>> np.round(qp1d.integrand(0.2), 2)
    0.04
    >>> qp1d.lower_bd
    0.41
    >>>
    >>> lower_bd = [0., 0.]
    >>> upper_bd = [1., 1.]
    >>> qp2d = QuadratureProblem(integrand, lower_bd=lower_bd, upper_bd=upper_bd)
    >>> qp2d.upper_bd
    [1.0, 1.0]
    """

    integrand: typing.Callable[[np.ndarray], typing.Union[float, np.ndarray]]
    lower_bd: typing.Union[pntp.FloatArgType, np.ndarray]
    upper_bd: typing.Union[pntp.FloatArgType, np.ndarray]
    output_dim: typing.Optional[int] = 1

    # For testing and benchmarking
    solution: typing.Optional[
        typing.Union[float, np.ndarray, pnrv.RandomVariable]
    ] = None



"""Iterated Gaussian filtering and smoothing."""
import numpy as np

import probnum.random_variables as pnrv

from .kalman import Kalman
from .stoppingcriterion import StoppingCriterion


class IteratedKalman(Kalman):
    """Iterated filter/smoother based on posterior linearisation.

    In principle, this is the same as a Kalman filter; however,
    1. predict() and update() in each filter step may be repeated.
    2. approximate Gaussian filtering and smoothing can use posterior linearisation.

    There is an additional method: :meth:`iterated_filtsmooth()`,
    which computes things like MAP estimates.
    """

    def __init__(self, kalman, stoppingcriterion=None):
        self.kalman = kalman
        if stoppingcriterion is None:
            self.stoppingcriterion = StoppingCriterion()
        else:
            self.stoppingcriterion = stoppingcriterion
        super().__init__(kalman.dynamics_model, kalman.measurement_model, kalman.initrv)

    def filter_step(
        self,
        start,
        stop,
        current_rv,
        data,
        previous_posterior=None,
        intermediate_step=None,
    ):
        """Filter step of iterated filter.

        Different to Kalman.filter in the sense that there may be multiple iterated updates for
        a single predict, or multiple iterated predicts for a single update, or multiple iterated predicts
        and updates in general.
        This retrieves methods such as the iterated extended Kalman filter.
        By further specifying a `previous_posterior` (KalmanPosterior), the first of those possibly
        iterated updates is linearised at the previous posterior estimate.
        This retrieves methods such as the iterated extended Kalman smoother.

        Parameters
        ----------
        start : float
            Predict FROM this time point.
        stop : float
            Predict TO this time point.
        current_rv : RandomVariable
            Predict based on this random variable. For instance, this can be the result
            of a previous call to filter_step.
        data : array_like
            Compute the update based on this data.
        previous_posterior : KalmanPosterior
            Posterior distribution of a previous smoothing iteration. Optional.
            If specified, posterior linearisation is applied.

        Returns
        -------
        RandomVariable
            Resulting filter estimate after the single step.
        dict
            Additional information provided by predict() and update().
            Contains keys `pred_rv`, `info_pred`, `meas_rv`, `info_upd`.
        """

        # initial prediction
        if previous_posterior is None:
            linearise_predict_at = None
            linearise_update_at = None
        else:
            linearise_predict_at = previous_posterior(start)
            linearise_update_at = previous_posterior(stop)
        pred_rv, info_pred = self.predict(
            start,
            stop,
            current_rv,
            linearise_at=linearise_predict_at,
            intermediate_step=intermediate_step,
        )
        upd_rv, meas_rv, info_upd = self.update(
            stop, pred_rv, data, linearise_at=linearise_update_at
        )

        # Specify additional information to be returned
        info = {
            "pred_rv": pred_rv,
            "info_pred": info_pred,
            "meas_rv": meas_rv,
            "info_upd": info_upd,
        }
        return upd_rv, info

    def iterated_filtsmooth(self, dataset, times, **kwargs):
        """Repeated filtering and smoothing using posterior linearisation."""
        posterior = self.filtsmooth(dataset, times, **kwargs)
        while self.stoppingcriterion.continue_filtsmooth_iteration(posterior):
            posterior = self.filter(dataset, times, linearise_at=posterior)
            posterior = self.smooth(posterior)
        return posterior

    def predict(self, start, stop, randvar, linearise_at=None, intermediate_step=None):
        """(Possibly iterated) prediction step."""
        pred_rv, info_pred = self.dynamics_model.transition_rv(
            randvar, start, stop=stop, linearise_at=linearise_at, step=intermediate_step
        )
        while self.stoppingcriterion.continue_predict_iteration(pred_rv, info_pred):
            pred_rv, info_pred = self.dynamics_model.transition_rv(
                pred_rv, start, stop, linearise_at=pred_rv, step=intermediate_step
            )
        return pred_rv, info_pred

    def update(self, time, randvar, data, linearise_at=None):
        """(Possibly iterated) update step."""
        upd_rv, meas_rv, info_upd = self._single_update(
            time, randvar, data, linearise_at=linearise_at
        )
        while self.stoppingcriterion.continue_update_iteration(
            upd_rv, meas_rv, info_upd
        ):
            upd_rv, meas_rv, info_upd = self._single_update(
                time, randvar, data, linearise_at=upd_rv
            )
        return upd_rv, meas_rv, info_upd

    def _single_update(self, time, randvar, data, linearise_at=None):
        # like kalman.update but with an explicit linearise_at argument
        meas_rv, info = self.measurement_model.transition_rv(
            randvar, time, linearise_at=linearise_at
        )
        crosscov = info["crosscov"]
        new_mean = randvar.mean + crosscov @ np.linalg.solve(
            meas_rv.cov, data - meas_rv.mean
        )
        new_cov = randvar.cov - crosscov @ np.linalg.solve(meas_rv.cov, crosscov.T)
        filt_rv = pnrv.Normal(new_mean, new_cov)
        return filt_rv, meas_rv, info



"""Bayesian Quadrature."""

from ._bayesquad import *
from .bq_methods import *

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "bayesquad",
    "BayesianQuadrature",
    "WarpedBayesianQuadrature",
]

# Set correct module paths. Corrects links and module paths in documentation.
BayesianQuadrature.__module__ = "probnum.quad"
WarpedBayesianQuadrature.__module__ = "probnum.quad"



"""Test problems from linear algebra."""

from ._random_spd_matrix import random_sparse_spd_matrix, random_spd_matrix

# Public classes and functions. Order is reflected in documentation.
__all__ = ["random_spd_matrix", "random_sparse_spd_matrix"]



from .extendedkalman import ContinuousEKFComponent, DiscreteEKFComponent
from .iterated_kalman import IteratedKalman
from .kalman import Kalman
from .kalmanposterior import KalmanPosterior
from .stoppingcriterion import FixedPointStopping, StoppingCriterion
from .unscentedkalman import ContinuousUKFComponent, DiscreteUKFComponent
from .unscentedtransform import UnscentedTransform



"""Stopping criteria for iterated filtering and smoothing."""
import numpy as np


class StoppingCriterion:
    """Stopping criteria for iterated filters/smoothers."""

    def __init__(
        self,
        max_num_predicts_per_step=None,
        max_num_updates_per_step=None,
        max_num_filtsmooth_iterations=None,
    ):
        self.num_predict_iterations = 0
        self.num_update_iterations = 0
        self.num_filtsmooth_iterations = 0

        self.max_num_predicts_per_step = max_num_predicts_per_step
        self.max_num_updates_per_step = max_num_updates_per_step
        self.max_num_filtsmooth_iterations = max_num_filtsmooth_iterations

    def continue_predict_iteration(self, pred_rv=None, info_pred=None):
        """Do we continue iterating the update step of the filter?"""
        return False

    def continue_update_iteration(self, upd_rv=None, meas_rv=None, info_upd=None):
        """Do we continue iterating the predict step of the filter?"""
        return False

    def continue_filtsmooth_iteration(self, kalman_posterior=None):
        """If implemented, iterated_filtsmooth() is unlocked."""
        raise NotImplementedError


class FixedPointStopping(StoppingCriterion):
    """Keep updating until the filter recursion arrives at a fixed-point."""

    def __init__(
        self,
        atol=1e-2,
        rtol=1e-2,
        max_num_predicts_per_step=1000,
        max_num_updates_per_step=1000,
        max_num_filtsmooth_iterations=1000,
    ):
        self.atol = atol
        self.rtol = rtol
        self.previous_pred_rv = None
        self.previous_upd_rv = None
        self.previous_posterior = None
        super().__init__(
            max_num_predicts_per_step=max_num_predicts_per_step,
            max_num_updates_per_step=max_num_updates_per_step,
            max_num_filtsmooth_iterations=max_num_filtsmooth_iterations,
        )

    def continue_predict_iteration(self, pred_rv=None, info_pred=None):
        """Do we continue iterating the update step of the filter?"""
        # Edge cases: zeroth or last iteration
        if self.num_predict_iterations >= self.max_num_predicts_per_step:
            raise RuntimeError("Maximum number of filter update iterations reached.")
        self.num_predict_iterations += 1
        if self.previous_pred_rv is None:
            self.previous_pred_rv = pred_rv
            return True

        # Compute relative thresholds
        mean_threshold = self.atol + self.rtol * np.maximum(
            np.abs(self.previous_pred_rv.mean), np.abs(pred_rv.mean)
        )
        cov_threshold = self.atol + self.rtol * np.maximum(
            np.abs(self.previous_pred_rv.cov), np.abs(pred_rv.cov)
        )

        # Accept if discrepancy sufficiently small
        mean_acceptable = np.all(
            np.abs(pred_rv.mean - self.previous_pred_rv.mean) < mean_threshold
        )
        cov_acceptable = np.all(
            np.abs(pred_rv.cov - self.previous_pred_rv.cov) < cov_threshold
        )
        continue_iteration = np.invert(
            np.all(np.logical_and(mean_acceptable, cov_acceptable))
        )
        if continue_iteration:
            self.previous_pred_rv = pred_rv
        return continue_iteration

    def continue_update_iteration(self, upd_rv=None, meas_rv=None, info_upd=None):
        """Do we continue iterating the predict step of the filter?"""
        # Edge cases: zeroth or last iteration
        if self.num_update_iterations >= self.max_num_updates_per_step:
            raise RuntimeError("Maximum number of filter update iterations reached.")
        self.num_update_iterations += 1
        if self.previous_upd_rv is None:
            self.previous_upd_rv = upd_rv
            return True

        # Compute relative thresholds
        mean_threshold = self.atol + self.rtol * np.maximum(
            np.abs(self.previous_upd_rv.mean), np.abs(upd_rv.mean)
        )
        cov_threshold = self.atol + self.rtol * np.maximum(
            np.abs(self.previous_upd_rv.cov), np.abs(upd_rv.cov)
        )

        # Accept if discrepancy sufficiently small
        mean_acceptable = np.all(
            np.abs(upd_rv.mean - self.previous_upd_rv.mean) < mean_threshold
        )
        cov_acceptable = np.all(
            np.abs(upd_rv.cov - self.previous_upd_rv.cov) < cov_threshold
        )
        continue_iteration = np.invert(
            np.all(np.logical_and(mean_acceptable, cov_acceptable))
        )
        if continue_iteration:
            self.previous_upd_rv = upd_rv
        return continue_iteration

    def continue_filtsmooth_iteration(self, kalman_posterior=None):

        # Edge cases: zeroth or last iteration
        if self.num_filtsmooth_iterations >= self.max_num_filtsmooth_iterations:
            raise RuntimeError("Maximum number of filter update iterations reached.")
        self.num_filtsmooth_iterations += 1
        if self.previous_posterior is None:
            self.previous_posterior = kalman_posterior
            return True

        # Compute relative thresholds
        mean_threshold = self.atol + self.rtol * np.maximum(
            np.abs(self.previous_posterior.state_rvs.mean),
            np.abs(kalman_posterior.state_rvs.mean),
        )
        cov_threshold = self.atol + self.rtol * np.maximum(
            np.abs(self.previous_posterior.state_rvs.cov),
            np.abs(kalman_posterior.state_rvs.cov),
        )

        # Accept if discrepancy sufficiently small
        mean_acceptable = np.all(
            np.abs(
                kalman_posterior.state_rvs.mean - self.previous_posterior.state_rvs.mean
            )
            < mean_threshold
        )
        cov_acceptable = np.all(
            np.abs(
                kalman_posterior.state_rvs.cov - self.previous_posterior.state_rvs.cov
            )
            < cov_threshold
        )
        continue_iteration = np.invert(
            np.all(np.logical_and(mean_acceptable, cov_acceptable))
        )
        if continue_iteration:
            self.previous_posterior = kalman_posterior
        return continue_iteration



"""Differential Equations."""

from .ode import (
    IVP,
    ODE,
    fitzhughnagumo,
    logistic,
    lotkavolterra,
    rigidbody,
    seir,
    threebody,
    vanderpol,
)
from .odefiltsmooth import GaussianIVPFilter, ivp2ekf0, ivp2ekf1, ivp2ukf, probsolve_ivp
from .odesolution import ODESolution
from .odesolver import ODESolver
from .steprule import AdaptiveSteps, ConstantSteps, StepRule

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "ODE",
    "IVP",
    "logistic",
    "fitzhughnagumo",
    "lotkavolterra",
    "seir",
    "rigidbody",
    "vanderpol",
    "threebody",
    "probsolve_ivp",
    "ODESolver",
    "GaussianIVPFilter",
    "ivp2ekf0",
    "ivp2ekf1",
    "ivp2ukf",
    "StepRule",
    "ConstantSteps",
    "AdaptiveSteps",
    "ODESolution",
]

# Set correct module paths (for superclasses).
# Corrects links and module paths in documentation.
ODE.__module__ = "probnum.diffeq"
ODESolver.__module__ = "probnum.diffeq"
StepRule.__module__ = "probnum.diffeq"



"""Markov transition rules: continuous and discrete."""

import abc
from typing import Dict, Optional

import numpy as np

from probnum.random_variables import RandomVariable

__all__ = ["Transition", "generate"]


class Transition(abc.ABC):
    """Markov transition rules in discrete or continuous time.

    In continuous time, this is a Markov process and described by a
    stochastic differential equation (SDE)

    .. math:: d x_t = f(t, x_t) d t + d w_t

    driven by a Wiener process :math:`w`. In discrete time, it is defined by
    a transformation

    .. math:: x_{t + \\Delta t} \\sim p(x_{t + \\Delta t}  | x_t).

    Sometimes, these can be equivalent. For example: mild solutions to
    linear, time-invariant SDEs have an equivalent, discretised form that can
    be written as a transformation.

    See Also
    --------
    :class:`ContinuousModel`
        Continuously indexed transitions (SDEs)
    :class:`DiscreteModel`
        Discretely indexed transitions (transformations)
    """

    def __init__(self):
        self.precon = None

    @abc.abstractmethod
    def transition_realization(
        self,
        real: np.ndarray,
        start: float,
        stop: Optional[float] = None,
        step: Optional[float] = None,
        linearise_at: Optional[RandomVariable] = None,
    ) -> (RandomVariable, Dict):
        """Transition a realization of a random variable from time :math:`t` to time
        :math:`t+\\Delta t`.

        For random variable :math:`x_t`, it returns the random variable defined by

        .. math:: x_{t + \\Delta t} \\sim p(x_{t + \\Delta t}  | x_t = r) .

        This is different to :meth:`transition_rv` which computes the parametrization
        of :math:`x_{t + \\Delta t}` based on the parametrization of :math:`x_t`.

        Nb: Think of transition as a verb, i.e. this method "transitions" a realization of a random variable.

        Parameters
        ----------
        real :
            Realization of the random variable.
        start :
            Starting point :math:`t`.
        stop :
            End point :math:`t + \\Delta t`.
        step :
            Intermediate step-size. Optional, default is None.
        linearise_at :
            For approximate transitions , for instance ContinuousEKFComponent,
            this argument overloads the state at which the Jacobian is computed.

        Returns
        -------
        RandomVariable
            Random variable, describing the state at time :math:`t + \\Delta t`
            based on realization at time :math:`t`.
        dict
            Additional information in form of a dictionary,
            for instance the cross-covariance in the
            prediction step, access to which is useful in smoothing.

        See Also
        --------
        :meth:`transition_rv`
            Apply transition to a random variable.
        """
        raise NotImplementedError

    def transition_realization_preconditioned(
        self,
        real: np.ndarray,
        start: float,
        stop: Optional[float] = None,
        step: Optional[float] = None,
        linearise_at: Optional[RandomVariable] = None,
    ) -> (RandomVariable, Dict):
        """Applies the transition, assuming that the state is already preconditioned.

        This is useful for numerically stable implementation of Kalman
        smoothing steps and Kalman updates.
        """
        if self.precon is None:
            errormsg = (
                "There is no preconditioned associated with this transition. "
                "Did you mean 'transition_realization'?"
            )
            raise NotImplementedError(errormsg)
        raise NotImplementedError(
            "'transition_realization_preconditioned' is not implemented."
        )

    @abc.abstractmethod
    def transition_rv(
        self,
        rv: "RandomVariable",
        start: float,
        stop: Optional[float] = None,
        step: Optional[float] = None,
        linearise_at: Optional[RandomVariable] = None,
    ) -> (RandomVariable, Dict):
        """Transition a random variable from time :math:`t` to time
        :math:`t+\\Delta t`.

        For random variable :math:`x_t`, it returns the random variable defined by

        .. math:: x_{t + \\Delta t} \\sim p(x_{t + \\Delta t}  | x_t) .

        This returns a random variable where the parametrization depends on the paramtrization of :math:`x_t`.
        This is different to :meth:`transition_rv` which computes the parametrization
        of :math:`x_{t + \\Delta t}` based on a realization of :math:`x_t`.

        Nb: Think of transition as a verb, i.e. this method "transitions" a random variable.


        Parameters
        ----------
        rv :
            Realization of the random variable.
        start :
            Starting point :math:`t`.
        stop :
            End point :math:`t + \\Delta t`.
        step :
            Intermediate step-size. Optional, default is None.
        linearise_at :
            For approximate transitions , for instance ContinuousEKFComponent,
            this argument overloads the state at which the Jacobian is computed.

        Returns
        -------
        RandomVariable
            Random variable, describing the state at time :math:`t + \\Delta t`
            based on realization at time :math:`t`.
        dict
            Additional information in form of a dictionary,
            for instance the cross-covariance in the
            prediction step, access to which is useful in smoothing.

        See Also
        --------
        :meth:`transition_realization`
            Apply transition to a realization of a random variable.
        """
        raise NotImplementedError

    def transition_rv_preconditioned(
        self,
        rv: "RandomVariable",
        start: float,
        stop: Optional[float] = None,
        step: Optional[float] = None,
        linearise_at: Optional[RandomVariable] = None,
    ) -> (RandomVariable, Dict):
        """Applies the transition, assuming that the state is already preconditioned.

        This is useful for numerically stable implementation of Kalman
        smoothing steps and Kalman updates.
        """
        if self.precon is None:
            errormsg = (
                "There is no preconditioned associated with this transition. "
                "Did you mean 'transition_rv'?"
            )
            raise NotImplementedError(errormsg)
        raise NotImplementedError("'transition_rv_preconditioned' is not implemented.")

    @property
    def dimension(self) -> int:
        """Dimension of the transition model.

        Not all transition models have a unique dimension. Some turn a
        state (x, y) into a scalar z and it is not clear whether the
        dimension should be 2 or 1.
        """
        raise NotImplementedError


def generate(dynmod, measmod, initrv, times, num_steps=5):
    """Samples true states and observations at pre-determined timesteps "times" for a
    state space model.

    Parameters
    ----------
    dynmod : statespace.Transition
        Transition model describing the prior dynamics.
    measmod : statespace.Transition
        Transition model describing the measurement model.
    initrv : probnum.RandomVariable object
        Random variable according to initial distribution
    times : np.ndarray, shape (n,)
        Timesteps on which the states are to be sampled.
    num_steps : int
        Number of steps to be taken for numerical integration
        of the continuous prior model. Optional. Default is 5.
        Irrelevant for time-invariant or discrete models.

    Returns
    -------
    states : np.ndarray; shape (len(times), dynmod.dimension)
        True states according to dynamic model.
    obs : np.ndarray; shape (len(times)-1, measmod.dimension)
        Observations according to measurement model.
    """
    states = np.zeros((len(times), _read_dimension(dynmod, initrv)))
    obs = np.zeros((len(times) - 1, _read_dimension(measmod, initrv)))
    states[0] = initrv.sample()
    for idx in range(1, len(times)):
        start, stop = times[idx - 1], times[idx]
        step = (stop - start) / num_steps
        next_state_rv, _ = dynmod.transition_realization(
            real=states[idx - 1], start=start, stop=stop, step=step
        )
        states[idx] = next_state_rv.sample()
        next_obs_rv, _ = measmod.transition_realization(real=states[idx], start=stop)
        obs[idx - 1] = next_obs_rv.sample()
    return states, obs


def _read_dimension(transition, initrv):
    """Extracts dimension of a transition without calling .dimension(), which is not
    implemented everywhere."""
    # relies on evaluating at zero, which is a dangerous endeavour and therefore,
    # this method is not used in Transition.dimension
    transitioned, _ = transition.transition_realization(
        real=initrv.mean, start=0.0, stop=1.0
    )
    return len(transitioned.sample())



"""Probabilistic State Space Models.

This package implements continuous-discrete and discrete-discrete state
space models, which are the basis for Bayesian filtering and smoothing,
but also probabilistic ODE solvers.
"""

from .discrete_transition import (
    DiscreteGaussian,
    DiscreteLinearGaussian,
    DiscreteLTIGaussian,
)
from .integrator import IBM, IOUP, Integrator, Matern
from .preconditioner import NordsieckLikeCoordinates, Preconditioner
from .sde import (
    LTISDE,
    SDE,
    LinearSDE,
    linear_sde_statistics,
    matrix_fraction_decomposition,
)
from .transition import Transition, generate

__all__ = [
    "Transition",
    "SDE",
    "LinearSDE",
    "LTISDE",
    "Integrator",
    "IBM",
    "IOUP",
    "Matern",
    "DiscreteGaussian",
    "DiscreteLinearGaussian",
    "DiscreteLTIGaussian",
    "Preconditioner",
    "NordsieckLikeCoordinates",
    "linear_sde_statistics",
    "matrix_fraction_decomposition",
    "generate",
]



"""Integrated processes such as the integrated Wiener process or the Matern process.

This is the sucessor of the former ODEPrior.
"""
try:
    # cached_property is only available in Python >=3.8
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

import numpy as np
import scipy.special

import probnum.random_variables as pnrv

from . import discrete_transition, sde
from .preconditioner import NordsieckLikeCoordinates


class Integrator:
    """An integrator is a special kind of SDE, where the :math:`i` th coordinate models
    the :math:`i` th derivative."""

    def __init__(self, ordint, spatialdim):
        self.ordint = ordint
        self.spatialdim = spatialdim

    def proj2coord(self, coord: int) -> np.ndarray:
        """Projection matrix to :math:`i` th coordinates.

        Computes the matrix

        .. math:: H_i = \\left[ I_d \\otimes e_i \\right] P^{-1},

        where :math:`e_i` is the :math:`i` th unit vector,
        that projects to the :math:`i` th coordinate of a vector.
        If the ODE is multidimensional, it projects to **each** of the
        :math:`i` th coordinates of each ODE dimension.

        Parameters
        ----------
        coord : int
            Coordinate index :math:`i` which to project to.
            Expected to be in range :math:`0 \\leq i \\leq q + 1`.

        Returns
        -------
        np.ndarray, shape=(d, d*(q+1))
            Projection matrix :math:`H_i`.
        """
        projvec1d = np.eye(self.ordint + 1)[:, coord]
        projmat1d = projvec1d.reshape((1, self.ordint + 1))
        projmat = np.kron(np.eye(self.spatialdim), projmat1d)
        return projmat


class IBM(Integrator, sde.LTISDE):
    """Integrated Brownian motion in :math:`d` dimensions."""

    def __init__(self, ordint, spatialdim, diffconst):
        self.diffconst = diffconst

        # initialise BOTH superclasses' inits.
        # I don't like it either, but it does the job.
        Integrator.__init__(self, ordint=ordint, spatialdim=spatialdim)
        sde.LTISDE.__init__(
            self,
            driftmat=self._driftmat,
            forcevec=self._forcevec,
            dispmat=self._dispmat,
        )

        self.precon = NordsieckLikeCoordinates.from_order(ordint, spatialdim)

    @property
    def _driftmat(self):
        driftmat_1d = np.diag(np.ones(self.ordint), 1)
        return np.kron(np.eye(self.spatialdim), driftmat_1d)

    @property
    def _forcevec(self):
        force_1d = np.zeros(self.ordint + 1)
        return np.kron(np.ones(self.spatialdim), force_1d)

    @property
    def _dispmat(self):
        dispmat_1d = np.zeros(self.ordint + 1)
        dispmat_1d[-1] = self.diffconst
        return np.kron(np.eye(self.spatialdim), dispmat_1d).T

    @cached_property
    def equivalent_discretisation_preconditioned(self):
        """Discretised IN THE PRECONDITIONED SPACE."""
        empty_force = np.zeros(self.spatialdim * (self.ordint + 1))
        return discrete_transition.DiscreteLTIGaussian(
            dynamicsmat=self._dynamicsmat,
            forcevec=empty_force,
            diffmat=self._diffmat,
        )

    @cached_property
    def _dynamicsmat(self):
        # Loop, but cached anyway
        driftmat_1d = np.array(
            [
                [
                    scipy.special.binom(self.ordint - i, self.ordint - j)
                    for j in range(0, self.ordint + 1)
                ]
                for i in range(0, self.ordint + 1)
            ]
        )
        return np.kron(np.eye(self.spatialdim), driftmat_1d)

    @cached_property
    def _diffmat(self):
        # Optimised with broadcasting
        range = np.arange(0, self.ordint + 1)
        denominators = 2.0 * self.ordint + 1.0 - range[:, None] - range[None, :]
        diffmat_1d = 1.0 / denominators
        return np.kron(np.eye(self.spatialdim), self.diffconst ** 2 * diffmat_1d)

    def transition_rv(self, rv, start, stop, **kwargs):
        if not isinstance(rv, pnrv.Normal):
            errormsg = (
                "Closed form transitions in LTI SDE models is only "
                "available for Gaussian initial conditions."
            )
            raise TypeError(errormsg)
        step = stop - start
        rv = self.precon.inverse(step) @ rv
        rv, info = self.transition_rv_preconditioned(rv, start)
        info["crosscov"] = self.precon(step) @ info["crosscov"] @ self.precon(step).T
        return self.precon(step) @ rv, info

    def transition_rv_preconditioned(self, rv, start, **kwargs):
        return self.equivalent_discretisation_preconditioned.transition_rv(rv, start)

    def transition_realization(self, real, start, stop, **kwargs):
        if not isinstance(real, np.ndarray):
            raise TypeError(f"Numpy array expected, {type(real)} received.")
        step = stop - start
        real = self.precon.inverse(step) @ real
        real, info = self.transition_realization_preconditioned(real, start)
        info["crosscov"] = self.precon(step) @ info["crosscov"] @ self.precon(step).T
        return self.precon(step) @ real, info

    def transition_realization_preconditioned(self, real, start, **kwargs):
        return self.equivalent_discretisation_preconditioned.transition_realization(
            real, start
        )

    def discretise(self, step):
        """Equivalent discretisation of the process.

        Overwrites matrix-fraction decomposition in the super-class.
        Only present for user's convenience and to maintain a clean
        interface. Not used for transition_rv, etc..
        """

        dynamicsmat = (
            self.precon(step)
            @ self.equivalent_discretisation_preconditioned.dynamicsmat
            @ self.precon.inverse(step)
        )
        diffmat = (
            self.precon(step)
            @ self.equivalent_discretisation_preconditioned.diffmat
            @ self.precon(step).T
        )
        zero_force = np.zeros(len(dynamicsmat))
        return discrete_transition.DiscreteLTIGaussian(
            dynamicsmat=dynamicsmat, forcevec=zero_force, diffmat=diffmat
        )


class IOUP(Integrator, sde.LTISDE):
    """Integrated Ornstein-Uhlenbeck process in :math:`d` dimensions."""

    def __init__(
        self,
        ordint: int,
        spatialdim: int,
        driftspeed: float,
        diffconst: float,
    ):
        # Other than previously in ProbNum, we do not use preconditioning for IOUP by default.
        self.driftspeed = driftspeed
        self.diffconst = diffconst

        Integrator.__init__(self, ordint=ordint, spatialdim=spatialdim)
        sde.LTISDE.__init__(
            self,
            driftmat=self._driftmat,
            forcevec=self._forcevec,
            dispmat=self._dispmat,
        )

    @property
    def _driftmat(self):
        driftmat_1d = np.diag(np.ones(self.ordint), 1)
        driftmat_1d[-1, -1] = -self.driftspeed
        return np.kron(np.eye(self.spatialdim), driftmat_1d)

    @property
    def _forcevec(self):
        force_1d = np.zeros(self.ordint + 1)
        return np.kron(np.ones(self.spatialdim), force_1d)

    @property
    def _dispmat(self):
        dispmat_1d = np.zeros(self.ordint + 1)
        dispmat_1d[-1] = self.diffconst
        return np.kron(np.eye(self.spatialdim), dispmat_1d).T


class Matern(Integrator, sde.LTISDE):
    """Matern process in :math:`d` dimensions."""

    def __init__(
        self,
        ordint: int,
        spatialdim: int,
        lengthscale: float,
        diffconst: float,
    ):

        # Other than previously in ProbNum, we do not use preconditioning for Matern by default.
        self.lengthscale = lengthscale
        self.diffconst = diffconst

        Integrator.__init__(self, ordint=ordint, spatialdim=spatialdim)
        sde.LTISDE.__init__(
            self,
            driftmat=self._driftmat,
            forcevec=self._forcevec,
            dispmat=self._dispmat,
        )

    @property
    def _driftmat(self):
        driftmat = np.diag(np.ones(self.ordint), 1)
        nu = self.ordint + 0.5
        D, lam = self.ordint + 1, np.sqrt(2 * nu) / self.lengthscale
        driftmat[-1, :] = np.array(
            [-scipy.special.binom(D, i) * lam ** (D - i) for i in range(D)]
        )
        return np.kron(np.eye(self.spatialdim), driftmat)

    @property
    def _forcevec(self):
        force_1d = np.zeros(self.ordint + 1)
        return np.kron(np.ones(self.spatialdim), force_1d)

    @property
    def _dispmat(self):
        dispmat_1d = np.zeros(self.ordint + 1)
        dispmat_1d[-1] = self.diffconst
        return np.kron(np.eye(self.spatialdim), dispmat_1d).T
